/**************************************************************************

        @file build.d

        Copyright (c) 2005 Derek Parnell

        This software is provided 'as-is', without any express or implied
        warranty. In no event will the authors be held liable for damages
        of any kind arising from the use of this software.

        Permission is hereby granted to anyone to use this software for any
        purpose, including commercial applications, and to alter it and/or
        redistribute it freely, subject to the following restrictions:

        1. The origin of this software must not be misrepresented; you must
           not claim that you wrote the original software. If you use this
           software in a product, an acknowledgment within documentation of
           said product would be appreciated but is not required.

        2. Altered source versions must be plainly marked as such, and must
           not be misrepresented as being the original software.

        3. This notice may not be removed or altered from any distribution
           of the source.

        4. Derivative works are permitted, but they must carry this notice
           in full and credit the original source.


                        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


        @version        Initial version, January 2005
        @author         Derek Parnell, Melbourne, Australia.


**************************************************************************/
/*
   A utility for building executables from source code written
   with the D programming language.

   This examines a set of D source code files to determine what
   to build and what needs to be included in the build. It then
   checks to see which components are current and creates a list
   of files that need to be compiled and/or included in the target
   file.

   This is based on the work called 'dmake v0.21' by Helmut Leitner
*/

/* --------- CHANGE LOG --------------------
__ /makedoc title Documentation for Build v2.09

__ /topic Change Log
__ /info
__ A list of changes and fixes that have been made.
__
__ /b"-- v2.09 -- 10/Aug/2005"
__ <ul>
__ /li ** /b FIX: /i"thanks to barrett9h (rodolfo)": Now correctly handles the return value
__ from Unix system() call.
__ /li ** /b FIX: /i"thanks to oniony": Now handles non_ASCII characters in PATH
__ environment symbol.
__ /li ** /b CHG: The "-silent" switch now also hides the linker stdout display.
__ /li ** /b CHG: Now supports and requires DMD v0.126 and GDC v0.13 or later.
__ /li ** /b ENH: New pragma(export_version) allows specified version identifiers to
__ be passed to all modules being compiled.
__ /li ** /b ENH: New command line switch "-run" to run the program
__ after a successful link.
__ </ul>
__
__ /b"-- v2.08 -- 29/May/2005"
__ <ul>
__ /li ** /b FIX: /i"thanks to teqdruid": In Unix environments,
__ any pragma(link, ~<name~>) statements were not sending the correct
__ syntax to the compiler's command line. The "~<name~>" is now prefixed
__ with "-L-l" for Unix systems.
__ /li ** /b FIX: /i"thanks to Carlos": In Unix environments, any "-version"
__ switch on the Utility's command line was not being converted to the correct Unix
__ format of "-fversion...". So now, if the Build tool gets
__ either "-version..." or "-fversion=..." on its command line, it will
__ recognise it as a 'Version' request for the compiler /b and output
__ it on the command line in the correct format for the environment it
__ is running in, namely "-fversion..." for GNU (gdc) and "-version..."
__ for Windows. /n
__ /b"Note:" The same applies to the "-debug..." and "-fdebug..." switches.
__ /li ** /b FIX: The pragma(nolink) was being ignored under some circumstances.
__ Now it splits the compilation phase
__ from the linking phase and excludes the 'nolink' files from the linker's
__ command line.
__ /li ** /b FIX: /i"thanks to kris": 'debug' statements were not being taken into account when
__ examining code for import statements.
__ /li ** /b FIX: 'debug' and 'version' levels are now being taken into account.
__ /li ** /b FIX: 'debug' and 'version' values being set inside a source file
__ were being made global rather than module scoped.
__ /li ** /b FIX: /i"thanks to carlos": The utility was not processing the
__ -I switch correctly when using GDC compiler.
__ /li ** /b ENH: The utility can now accept the -I switch in two forms:
__ /b"-I~<path~>" and /b"-I ~<path~>" on its command line, regardless of which
__ compiler is being used. When invoking the compiler, it uses the correct
__ format on the command line of the compiler being used, namely "-I~<path~>" to
__ DMD compiler and "-I ~<path~>" to GDC.
__ /li ** /b ENH: The path of any source file that imports a module
__ will be added to the list of root paths to search for module source
__ files. This means that you can now have module source files referenced
__ relative to the source file that imports them. This is the new default
__ behaviour but it can be turned off by using the new -noautoimport switch
__ or for individual files by placing them inside parenthesis.
__ /li ** /b ENH: New switch /i"-noautoimport" is used to turn off the
__ automatic adding of search roots based on the path of the source files
__ being compiled.
__ /li ** /b ENH: New switch /i"-LIBPATH=" is used to add search paths for
__ library files.
__ /li ** /b WARN: The utility only does a single scan of each source file
__ and thus if an file sets a 'debug' or 'version' value after a function
__ which uses that value, it will be ignored. You are requested to ensure that
__ all your 'debug' and 'version' setting is done prior to the first module
__ member.
__ </ul>
__
__ /b"-- v2.07 -- 06/May/2005"
__ <ul>
__ /li ** /b ENH: New switch /i"-nodef" to prevent the utility from
__ automatically creating a Module Definition File (.def)
__ /li ** /b ENH: You can now specify default, and special, settings for
__ the utility's command line. These are placed in a text configuration file called
__ /b"build.cfg". The utility first looks in the same directory as the
__ utility's executable for the configuration file. After processing any
__ configuration file found there, it then looks in the same directory as
__ the compiler for a configuration file (build.cfg) and processes it if
__ found.
__ /li ** /b ENH: To support special command line settings that may be
__ specified in the utility's configuration file(s), you can indicate
__ one or more setting groups on the command line. These take the format of
__ +groupname. The configuration file(s) are scanned for special group settings
__ after the default ones have been processed.
__ /li ** /b ENH: To remove an earlier specified command line switch, you
__ can prepend it with a hyphen. This new feature may be needed when command
__ line switches can come from multiple sources (the original command line,
__ utility response files, and utility configuration files), and you need
__ to remove a switch that may have been provided by some other source. /n
__ For example to negate an earlier "-unittest" switch you code add "--unittest"
__ to the command line.
__ /li ** /b ENH: New switch /i"-od" is used to nominate a directory for
__ temporary files. For the DigitalMars compiler, this also is used as the location
__ to create object files.
__ </ul>
__
__  /b"-- v2.06 -- 04/May/2005"
__ <ul>
__ /li ** /b FIX: /i"thanks to phoenix": When emitting commandline parameters for
__ the compiler or librarian, it was possible to have some ending with /b'\"' which caused
__ the shell to /i escape the quote character and thus provide incorrect parameters to
__ the compiler.
__ </ul>
__  /b"-- v2.05 -- 02/May/2005"
__ <ul>
__ /li ** /b FIX: When supplying some forms of relative paths, the
__ utility would crash or give the wrong canonical form of the path. This
__ effected formats like \..\anything and  C:..\anything
__ /li ** /b FIX: The utility now assumes the current directory if it can not
__ locate the compiler using the paths on the system PATH symbol.
__ /li ** /b FIX: The utility now assumes the compiler's location if no path
__ for the configuration file has been specified.
__ /li ** /b FIX: The utility now supports better parsing of the lines in the
__ configuration file. It handles odd variations of embedded quotes.
__ /li ** /b ENH: The utility now supports Windows 95/98/ME for file times.
__ </ul>
__
__  /b"-- v2.04 -- 29/Apr/2005"
__ <ul>
__ /li ** /b FIX: /i"thanks to Nils Hensel": The utility was not correctly
__ parsing the DFLAGS line in the configuration file when it contained
__ embedded quote characters.
__ /li ** /b FIX: /i"thanks to Anders F Bjoerklund": The utility was not treating
__ directory names that contained dots correctly.
__ /li ** /b FIX: /i"thanks to Carlos": Unix-style files that end with a nested comment
__ delimiter are now handled correctly.
__ /li ** /b ENH: /i"thanks to qbert(Charlie)": The location of the librarian tool can now
__ be specified. It can be explictly named on a LIBCMD= line inside the
__ DMD configuration file, or failing that, implictly assumed to be in the
__ same directory as the DMD linker.
__ /li ** /b ENH: A new commandline switch -LIBOPT which allows you to pass
__ commandline options to the librarian.
__ </ul>
__
__  /b"-- v2.03 -- 20/Apr/2005"
__ <ul>
__ /li ** /b FIX: /i"thanks to Carlos": The linux build had a spelling mistake
__ in source.d (line 286). The indentifier "lNextMod" should have been coded
__  as "lNextModule".
__ </ul>
__
__  /b"-- v2.02 -- 19/Apr/2005"
__ <ul>
__ /li ** /b FIX: /i"thanks to Carlos": The utility was not handling the situation
__ where a simple statement, rather than a block statement, followed a version
__ directive.
__ /li ** /b FIX: /i"thanks to aldacron": The utility now respects the compiler's
__ 'Output file' switch. Which is '-of[filename]' for DMD and '-o [filename]'
__ for GNU. The utility treats it as an alias for its -T switch.
__ /li ** /b FIX: /i"thanks to aldacron": The utility now respects the DMD compiler's
__ 'Object Path' switch. If you have '-odsomepath' on the command line, the
__ utility adds this to the search path when checking for up-to-date files,
__ and will clean up files in that directory if requested to (-clean). It will
__ also create the directory if it doesn't exist.
__ </ul>
__
__  /b"-- v2.01 -- 18/Apr/2005"
__ <ul>
__ /li ** /b FIX: /i"thanks to Justin (jcc7)": If an excluded module (-X) was
__ in a package at the current directory level, it was not being excluded.
__ </ul>
__
__  /b"-- v2.00 -- 08/Apr/2005"
__ <ul>
__ /li ** /b FIX: Now supports the raw-string delimiter ` in Build's own
__ pragma commands.
__ /li ** /b FIX: Now handles backslash escapes in Build's own
__ pragma commands.
__ /li ** /b ENH: New -silent switch avoids all unnecessary messages.
__ /li ** /b ENH: New -help switch displays full usage text. Aliases: -h, -?
__ </ul>
__
__  /b"-- v1.19 -- 04/Apr/2005"
__ <ul>
__ /li ** /b FIX: /i"thanks to Ben Hinkle". When creating a library in Linux,
__ the utility was looking for /i".obj" files instead of /i".o" files.
__ /li ** /b FIX" When creating a library, if a source file had more than one
__ /i".d" in its name, the utility would look for the wrong object file for it.
__ </ul>
__
__  /b"-- v1.18 -- 03/Apr/2005"
__ <ul>
__ /li ** /b FIX: If mixed case was used, it was possible to have duplicate
__ element types in a OptLink module definition file. Effected the use of
__ pragma(build_def).
__ /li ** /b ENH: Some performance improvements.
__ </ul>
__
__  /b"-- v1.17 -- 30/Mar/2005"
__ <ul>
__ /li ** /b FIX: In Windows systems, the utility now does case-insensitive
__  path name compares.
__ /li ** /b FIX: A bug in the String Search function crashed the utility
__ when the -T switch was used with target names less than 9 characters long.
__ /li ** /b ENH: The new switch /b"-obj" is available. This is just a shorthand
__ for having both /i"-nolib" and /i"-nolink" on the command line.
__ </ul>
__
__  /b"-- v1.16 -- 28/Mar/2005"
__ <ul>
__ /li ** /b FIX: The utility now handles quotes around pragma(link) references.
__ /li ** /b FIX: The utility now handles multiple references to the same module
__ even though they are using different 'path' specifications.
__ /li ** /b ENH: The utility now has support for making Windows DLL files.
__ </ul>
__
__  /b"-- v1.15 -- 24/Mar/2005"
__ <ul>
__ /li ** /b FIX: The utility now checks for more recent libraries and
__ object files as added by pragma(link), pragma(build), and command line. It
__ didn't used to rebuild the executable if only a new version of a library
__ was present.
__ /li ** /b FIX: The order of the librarian parameters was incorrect.
__ /li ** /b FIX: The utility was not using the correct path name if importing
__ modules from some Import paths.
__ /li ** /b ENH: Verbose mode now shows the ignored and noticed packages.
__ </ul>
__
__  /b"-- v1.14 -- 23/Mar/2005"
__ <ul>
__ /li ** /b FIX: /i"Thanks to teqDruid." For GNU platforms, the utility
__  now passes any library name found in pragma(link,...) to the linker.
__ /li ** /b FIX: Nows correctly handles unix and Microsoft line-ends.
__ /li ** /b FIX: /i"Thanks to Ben Hinkle." Using relative paths '..' and '.'
__ now work correctly on the -I switch.
__ /li ** /b FIX: /i"Thanks to Ben Hinkle." The correct librarian program 'ar'
__ is now called on unix systems.
__ /li ** /b ENH: The new switch -names displays the names of the files
__ used in building the target. This is not a noisy as verbose mode.
__ /li ** /b ENH: New pragma /b include used to identify a file that
__ must be included in the compilation process but is not one that
__ is imported by any file in the group. Only needed if command line
__  file does not otherwise import the required file. Can be used to
__  include the file containing the 'main' function from a file that
__ does not import that file. Rarely needed.
__ /li ** /b ENH: New pragma /b build used to build 'foreign' file types
__ that D doesn't know about.
__ /li ** /b ENH: New pragma /b target used to nominate the default target
__ file name.
__ /li ** /b ENH: You can now use a symbolic target name on the -T switch.
__ </ul>
__
__  /b"-- v1.13 -- 13/Mar/2005"
__ <ul>
__ /li ** /b FIX: Explictly included paths entered on the -I switch are
__ now added to the dmd command line.
__ /li ** /b FIX: Explictly excluded modules (via -X) were only checking package
__ names. They also check for module names now.
__ /li ** /b FIX: Imported files are now looked for relative to the current directory
__ before the import paths are checked.
__ /li ** /b FIX: The utility now checks the DFLAG environment symbol in
__ non-DigitalMars D compilers.
__ /li ** /b CHG: The switch -nounittest is now deprecated and ignored if used.
__ The default setting is no longer to compile with unittests turned on.
__ </ul>
__  /b"-- v1.12 -- 6/Mar/2005"
__ <ul>
__ /li ** /b FIX: The response file (-Ry) is now only the default when using
__ DigitalMars tools on Windows. All other environments do not use the response
__ file by default.
__ </ul>
__  /b"-- v1.11 -- 4/Mar/2005"
__ <ul>
__ /li ** /b FIX: No long outputs empty compiler tool switches.
__ </ul>
__
__  /b"-- v1.10 -- 4/Mar/2005"
__ <ul>
__ /li ** /b FIX: /i"Thanks to Carlos." Better support for GDC command line switches.
__ /li ** /b FIX: /i"Thanks to Carlos." Cause 'Unix' version to trigger Posix code.
__ /li ** /b FIX: /i"Thanks to Carlos." The vUseResponseFile variable was only being
__ declared in the Windows version.
__ /li ** /b FIX: No passes space-embedded switches and paths to the compiler
__ tools correctly.
__ /li ** /b ENH: If an environment symbol called DFLAGS has been defined, it is
__ analyzed. This is done before any configuration file processing.
__ /li ** /b ENH: New command line switch -test. This shows the command lines
__ instead of running them. No compiling or linking is done.
__ </ul>
__
__  /b"-- v1.9 -- 3/Mar/2005"
__ <ul>
__ /li ** /b FIX: /i"Thanks to Carlos." For unix, it now uses ':' instead of ';' for path
__ separators when parsing the configuration file lines.
__ /li ** /b FIX: /i"Thanks to Carlos." The default location for the configuration
__ file was being ignored.
__ /li ** /b ENH: A new command line switch /b"-R" used to control the use
__  of a response file for the compiler tools.
__ </ul>
__
__  /b"-- v1.8 -- 1/Mar/2005"
__ <ul>
__ /li ** /b FIX: /i"Thanks to Anders F Bjoerklund:" Removed nonASCII chars from source code.
__ /li ** /b FIX: /i"Thanks to Anders F Bjoerklund:" Changed version(GDC) to version(GNU)
__ /li ** /b FIX: Changed compiler name for GNU versions from 'dmd' to 'gdc'
__ /li ** /b FIX: /i"Thanks to Carlos." The application was incorrectly parsing the DFLAGS sc.ini line.
__ /li ** /b ENH: /i"Thanks to Anders F Bjoerklund:" Provided a Makefile for the initial
__ compile of the application.
__ </ul>
__
__  /b"-- v1.7 -- 28/Feb/2005"
__ <ul>
__ /li ** /b FIX: /i"Thanks to clayasaurus." Linux edition was not compiling
__ due to a missing 'version(Windows)' block around the use of vWinVer
__ </ul>
__
__  /b"-- v1.6 -- 28/Feb/2005"
__ <ul>
__ /li ** /b FIX: /i"Thanks to Kris." The utility was not including the DCPATH value
__ when invoking the compiler.
__ /li ** /b ENH: When building a Windows app, the .DEF file now includes the
__ version of windows that /i this application is being run on.
__ /li ** /b ENH: The command line switch /b"-gui" can now have an optional
__  version information to specify which Windows version to build for.
__ The default Windows Version used is the one for the version running
__  this utility.
__ /li ** /b ENH: New command line switch -info to display the utility's version #.
__ </ul>
__  /b"-- v1.5 -- 25/Feb/2005"
__ <ul>
__ /li ** /b FIX: /i"Thanks to clayasaurus." Now only displays the CFPATH
__  and DCPATH messages when /i verbose mode is on.
__ /li ** /b FIX: /i"Thanks to kris." Hardcoded default for DMD path was
__ misleading when dmd.exe was not in the Windows PATH.
__ /li ** /b FIX: The utility was not correctly parsing empty raw strings ``.
__ /li ** /b ENH: /i"Thanks to kris." If the -DCPATH switch is used, then
__ if the existing DCPATH and CFPATH are the same, this switch changes both.
__ </ul>
__
__  /b"-- v1.4 -- 24/Feb/2005"
__ <ul>
__ /li ** /b FIX: /i"Thanks to Abscissa." The -version switch was being interpreted as
__  a file to compile.
__ /li ** /b FIX: /i"Thanks to brad." Linux version was using the wrong package name
__ when formatting a file date for display.
__ /li ** /b FIX: /i"Thanks to brad." Linux version was not setting the target
__ name correctly which resulted in the first source file being compiled twice.
__ /li ** /b FIX: /i"Thanks to brad." Parsing the DFLAG line in the dmd options
__ file was not dealing with whitespace delimiters.
__ /li ** /b ENH: When specifing an alternate target file, you can now also
__  specify a path for the target, and that path is created for you if it
__ doesn't exist.
__ </ul>
__
__  /b"-- v1.3 -- 23/Feb/2005"
__ <ul>
__ /li ** /b ENH: Put in a workaround for DMD (Windows) not using the sc.ini
__ file correctly. The utility now explicitly adds any DFLAG options from
__ the sc.ini file into the .rsp file.
__ /li ** /b FIX: The utility now correctly handles ';' delimited DFLAG options
__ in the sc.ini file.
__ /li ** /b ENH: Now supports the use of 'response' files on the command line
__ /li ** /b ENH: Now supports the -T command line switch to supply an
__  alternate tareget name.
__
__ </ul>
__  /b"-- v1.2 -- 21/Feb/2005"
__ <ul>
__ /li ** /b ENH: /i"Thanks to Anders F Bjoerklund:" Added better support for 'darwin'
__        and 'gdc'.
__ </ul>
*/

module build;
private import build_bn;    // This module's build number

version(unix)  version = Unix;
version(Unix)  version = Posix;
version(linux)  version = Posix;
version(darwin) version = Posix;

version(build)
{
    version(Windows) {
        // OptLink Definition File
        pragma (build_def, "VERSION 2.9");
    }
}


private{
    // --------- imports ----------------
    import source;         // Source File class

    import util.str;       // non-standard string routines.
    import util.fdt;       // File Date-Time class
    import util.pathex;    // Extended Path routines.
    import util.fileex;    // Extended File routines.
    import util.macro;     // Macro processing routines.

    import std.c.stdio;
    import std.file;
    import std.outbuffer;
    import std.path;
    import std.stdio;
    import std.stream;
    import std.string;

    version(Windows) {
        import std.c.windows.windows;
    }
    else version(linux) {
        import std.c.linux.linux;
    }
    else version(darwin) {
        import std.c.darwin.darwin;
    }

    // --------- C externals ----------------
    extern (C)
    {
        char*   getenv  (char *);
        int     putenv  (char *);
        int     system  (char *);
    }


    // --------- enums ----------------
    enum LibOpt {Implicit, Build, DontBuild}

    // --------- constants ----------------
    const {
        version(Windows) {
            char[] kExeExtention=`exe`;
            char[] kLibExtention=`lib`;
            char[] kObjExtention=`obj`;
            char[] kShrLibExtention=`dll`;
        }

        version(Posix) {
            char[] kExeExtention=``;
            char[] kLibExtention=`a`;
            char[] kObjExtention=`o`;
            char[] kShrLibExtention=`s`;
        }
        char[] kSrcExtention=`d`;
        char[] kMacroExtention=`mac`;

    }

    // ---------- Module scoped globals -----------
    version(DigitalMars) {
        version(Windows) {
            char[] vCompiler=`dmd.exe`;
            char[] vCompilerSwitch= ` -c`;
            char[] vLinker=`dmd.exe`;
            char[] vConfigFile=`sc.ini`;
            char[] vCompilerPath=``;
            char[] vConfigPath=``;
            char[] vLibPaths = ``;
            char[] vConfigSep = ";";
            char[] vLibrarian = `lib`;
            char[] vLibrarianOpts = `-c -p256`;
        }

        version(Posix) {
            char[] vCompiler=`dmd`;
            char[] vCompilerSwitch= ` -c`;
            char[] vLinker=`dmd`;
            char[] vConfigFile=`dmd.conf`;
            char[] vCompilerPath=``;
            char[] vConfigPath=`/etc/`;
            char[] vLibPaths = ``;
            char[] vConfigSep = ":";
            char[] vLibrarian = `ar`;
            char[] vLibrarianOpts = `-r`;
        }

        char[]     vVersionSwitch = "-version=";
        char[]     vDebugSwitch = "-debug";
        char[]     vKeepPathSwitch = "-op";
        char[]     vOutFileSwitch = "-of";
        char[]     vImportPathSwitch = "-I";
        char[]     vNoLinkSwitch    = "-c";
    }

    version(GNU) {
        version(Windows) {
            char[] vCompiler=`gdc.exe`;
            char[] vCompilerSwitch= ` -c`;
            char[] vLinker=`gdc.exe`;
            char[] vConfigFile=null;
            char[] vCompilerPath=``;
            char[] vConfigPath=null;
            char[] vLibPaths = ``;
            char[] vConfigSep = ";";
            char[] vLibrarian = `lib`;
            char[] vLibrarianOpts = `-c`;
        }

        version(Posix) {
            char[] vCompiler=`gdc`;
            char[] vCompilerSwitch= ` -c`;
            char[] vLinker=`gdc`;
            char[] vConfigFile=null;
            char[] vCompilerPath=`/opt/gdc/bin`;
            char[] vConfigPath=null;
            char[] vLibPaths = ``;
            char[] vConfigSep = ":";
            char[] vLibrarian = `ar`;
            char[] vLibrarianOpts = `-r`;
        }
        char[]     vVersionSwitch = "-fversion=";
        char[]     vDebugSwitch = "-fdebug";
        char[]     vOutFileSwitch = "-o ";
        char[]     vKeepPathSwitch = "";
        char[]     vImportPathSwitch = "-I ";
        char[]     vNoLinkSwitch    = "-c";
    }

    char[]       vBuildImportPathSwitch = "-I";
    char[]       vTemporaryPathSwitch = "-od";
    char[]       vRunSwitch = "-run";
    char[]       vLibrarianPath = "";
    char[]*      vDelayedValue = null;
    char[]       vTemporaryPath = "";
    bool         vTestRun = false;
    bool         vNoLink = false;
    bool         vForceCompile = false;
    bool         vSilent = false;
    bool         vCleanup = false;
    bool         vVerbose = false;
    bool         vNames = false;
    bool         vAllObjects = false;
    bool         vNoDef = false;
    bool         vAutoImports = true;
    bool         vMacroInput = true;
    bool         vExecuteProgram = false;
    char[]       vRunParms;
    char[]       vTargetExe;
    char[][]     vImportRoots;
    char[][]     vModulesToIgnore;
    char[][]     vModulesToNotice;
    char[][]     vBuildDef;
    char[][]     vDefaultLibs;
    char[][]     vDefaultCompArgs;
    LibOpt       vLibraryAction = LibOpt.Implicit;
    char[]       vAppPath;
    char[]       vAppName;
    char[]       vAppVersion = "2.9";
    char[]       vTargetName;           // Output name from first file name.
    char[]       vPragmaTargetName;     // Output name from pragma.
    char[]       vCommandTargetName;    // Output name from switches.
    char[][]     vCmdLineSourceFiles;   // List of source files from command line
    bool[char[]] vLinkFiles;            // List of non-source files from command line
    char[][]     vCombinedArgs;         // All the args are gathered here prior to processing.
    char[][]     vBuildArgs;            // Arguments passed to build
    char[][]     vCompilerArgs;         // Arguments passed to compiler
    bool         vUseResponseFile;
    bool         vConsoleApp = true;
    char[]       vRDFName = "default.rdf";
    bool[char[]] vNestedDelim;
    struct ExternRef
    {
        char[]   FilePath;
        char[][] ToolOpts;
    }
    ExternRef[]  vExternals;

    Substitute[] vSubstitutions;

 version(Windows) {
    char[]       vWinVer = "";
    ubyte        vWinVerNum;
    }

}

// Module constructor.
//-------------------------------------------------------
static this()
//-------------------------------------------------------
{
    source.ActivateVersion("build");

    vDefaultCompArgs ~= vKeepPathSwitch;     // Keep object files in same folder as source.
    vModulesToIgnore ~= "phobos";  // Assume phobos will be ignored.

    version(Posix){
        vDefaultCompArgs ~= vVersionSwitch ~ "Posix"; // Until such time as this is standard in dmd.
    }

    version(Windows) {
        vWinVerNum = cast(ubyte)(GetVersion() & 0xFF);
        vWinVer = std.string.format("%d.0", vWinVerNum);
     }

    version(DigitalMars) {
        version(Windows) {
            vUseResponseFile = true;
        }
        else
        {
            vUseResponseFile = false;
        }
     } else
     {
        vUseResponseFile = false;
     }

    vNestedDelim["("] = true;
    vNestedDelim["<"] = true;
    vNestedDelim["{"] = true;
    vNestedDelim["["] = true;
}


//-------------------------------------------------------
void DisplayUsage(bool pFull = true)
//-------------------------------------------------------
{

    writefln(
        "Path and Version : %s v%s(%d)\n  built on %s"
            ,vAppPath, vAppVersion, build_bn.auto_build_number,
            __TIMESTAMP__);
    if (pFull == false)
        return;
    else {

    writefln(
        "Usage: %s sourcefile [options objectfiles libraries]"
            , vAppName);
    writefln("  sourcefile D source file");
    writefln("  -v         Verbose (passed through to D)");
    writefln("  -V         Verbose (NOT passed through)");
    writefln("  -names     Displays the names of the files used in building the target.");
    writefln("  -DCPATH<path> <path> is where the compiler has been installed.");
    writefln("             Only needed if the compiler is not in the system's");
    writefln("             PATH list. Used if you are testing an alternate");
    writefln("             version of the compiler.");
    writefln("  -CFPATH<path> <path> is where the D config file has been installed.");
    writefln("  -full      Causes all source files, except ignored modules,");
    writefln("              to be compiled.");
    writefln("  -link      Forces the linker to be called instead of the librarian.");
    writefln("              (Only needed if the source files do not contain");
    writefln("               main/WinMain)");
    writefln("  -nolink    Ensures that the linker is not called.");
    writefln("              (Only needed if main/WinMain is found in the source");
    writefln("               files and you do NOT want an executable created.)");
    writefln("  -lib       Forces the object files to be placed in a library.");
    writefln("              (Only needed if main/WinMain is found in the source");
    writefln("               files AND you want it in a library instead of");
    writefln("               an executable.)");
    writefln("  -nolib     Ensures that the object files are not used to form");
    writefln("              a library.");
    writefln("              (Only needed if main/WinMain is not found in the source");
    writefln("               files and you do NOT want a library.");
    writefln("  -obj       This is the same as having both -nolib and -nolink switches.");
    writefln("  -allobj    Ensures that all object files are added to a");
    writefln("              library.");
    writefln("              (Normally only those in the same directory are added.)");
    writefln("  -cleanup   Ensures that all object files created during the run");
    writefln("              are removed at the end of the run, plus other work files.");
    writefln("  -clean     Same as -cleanup");
  version(Windows) {
    writefln("  -gui[:x.y] Forces a GUI application to be created. The optional");
    writefln("              :x.y can be used to build an application for a ");
    writefln("              specific version of Windows. eg. -gui:4.0");
    writefln("              (Only needed if WinMain is not found in the source files");
    writefln("               or if you wish to override the default Windows version)");
    writefln("  -dll       Forces a DLL library to be created.");
    writefln("              (Only needed if DllMain is not found in the source files.)");
   }

    writefln("  -LIBOPT<opt> Allows you to pass <opt> to the librarian.");
    writefln("  -LIBPATH=<pathlist> Used to add a semi-colon delimited list");
    writefln("                of search paths for library files.");
    writefln("  -test      Does everything as normal except it displays the commands");
    writefln("              instead of running them.");
    writefln("  -RDF<path> Overrides the default Rule Definition File");
    writefln("  -R<y|n>    Indicates whether to use a response file or command line");
    writefln("              arguments with the compiler tools.");
    writefln("               -Ry will cause a response to be used.");
    writefln("               -Rn will cause command line arguments to be used.");
    writefln("               -R will reverse the current usage.");

  version(DigitalMars) {
   version(Windows) {
    writefln("               ** The default is to use a response file"); }
   else {
    writefln("               ** The default is to use command line arguments"); }
  }
  else {
    writefln("               ** The default is to use command line arguments"); }

    writefln("  -run<param> If the link is successful, this will cause the");
    writefln("               executable just created to run. You can give it ");
    writefln("               run time parameters. Anything after the '-run' will");
    writefln("               placed in the program's command line. You will need");
    writefln("               to quote any embedded spaces.");
    writefln("  -od<path>  Nominate the directory where temporary (work) files");
    writefln("             are to be created. By default they are created in");
    writefln("             the same directory as the target file.");
    writefln("  -X<module> Modules/Packages to ignore (eg. -Xmylib)");
    writefln("  -M<module> Modules/Packages to notice (eg. -Mphobos)");
    writefln("  -T<targetname> The name of the target file to create. Normally");
    writefln("              the target name istaken from the first or only name");
    writefln("              of the command line.");
    writefln("  -help     Displays the full 'usage' help text. ");
    writefln("  -h        Same as -help, displays the full 'usage' help text.");
    writefln("  -?        Same as -help, displays the full 'usage' help text.");
    writefln("  -silent   Avoids unnecessary messages being displayed.");
    writefln("  -noautoimport Turns off the automatic addition of source paths");
    writefln("              to the list of Import Roots.");
    writefln("  -info      Displays the version and path of the Build application.");
    writefln("  -nodef    Prevents a Module Definition File from being created.");
    writefln("  [...]      All other options, objectfiles and libraries are");
    writefln("              passed to the compiler");
    writefln("*Note, you can specify all or any command line value in a ");
    writefln("   response file. Each value appears in its own line in the");
    writefln("   response file and you reference this file by prefixing");
    writefln("   its name with an '@' symbol on the command line.");
    writefln("   Example:  build @final");
    writefln("      where a file called 'final.brf' contains the command");
    writefln("      line values (including other response file references)");
    writefln("   If the response file reference is just a single '@' then");
    writefln("   build looks for a file called 'build.brf'");
}
}

//-------------------------------------------------------
int RunCommand(char[] pCommand)
//-------------------------------------------------------
{
    int lRC;
    int lTrueRC;

    if (vTestRun) {
        writefln("Command: '%s'",pCommand);
        return 0;
    }
    else
    {


        if(vVerbose) {
            writefln("Running '%s'",pCommand);
        }

        lRC = system(std.string.toStringz(pCommand));
        version(Posix) lTrueRC = ((lRC & 0xFF00) >> 8);
        version(Windows) lTrueRC = lRC;

        if(vVerbose) {
            if (lTrueRC == 0){
                writefln("Successful");
            } else {
                writefln("Failed. Return code: %04x",lRC);
            }
        }
        return lTrueRC;
    }
}

//-------------------------------------------------------
char[] GetEnv(char[] pSymbol)
//-------------------------------------------------------
{
    return std.string.toString(getenv(pSymbol));
}

// Scans all known source files and extacts any modules.
// It returns the time of the most recently modified file.
//-------------------------------------------------------
FileDateTime GetNewestDateTime()
//-------------------------------------------------------
{
    Source lSource;
    FileDateTime lModsTime = new FileDateTime();

    foreach(int i, char[] lFileName; Source.SourceIndex.keys)
    {
        // Get the next Source object to examine.
        lSource = Source.SourceIndex[lFileName];
        if (lSource.Ignore)
            continue;

        // Ensure that it has been processed.
        lSource.search();

        if(vVerbose) {
            writefln("source file[%d] %s", i,
                    util.pathex.AbbreviateFileName(lFileName));
        }
        else if(vNames) {
            writefln(" [ %s ]", util.pathex.AbbreviateFileName(lFileName));
        }

        if (lSource.DependantsTime > lModsTime)
        {
            if(vVerbose) {
                writefln("Newer time: from %s to %s",
                    lModsTime.toString(),
                    lSource.DependantsTime.toString()
                    );
            }

            lModsTime = lSource.DependantsTime;
        }
    }

    // Examine any link file dependancies too.
    foreach(int i, char[] lFileName; vLinkFiles.keys)
    {
        FileDateTime lLinkTime = GetFileTime(lFileName);

        if(vVerbose) {
            writefln("link file[%d] %s %s", i,
                    util.pathex.AbbreviateFileName(lFileName), lLinkTime.toString());
        }
        else if(vNames) {
            writefln(" [ %s ]", util.pathex.AbbreviateFileName(lFileName));
        }

        if (lLinkTime > lModsTime)
        {
            if(vVerbose) {
                writefln("Newer time: from %s to %s",
                    lModsTime.toString(),
                    lLinkTime.toString()
                    );
            }

            lModsTime = lLinkTime;
        }
    }

    return lModsTime;
}

struct Rule
{
    char[] Name;
    char[] Input;
    char[] Output;
    char[] Tool;
};

//-------------------------------------------------------
Rule[] LoadRules()
//-------------------------------------------------------
{
    Rule[] lRules;
    char[][] lRuleText;
    char[] lRuleDefnFile;


    lRuleDefnFile = util.pathex.LocateFile(vRDFName,
                        getDirName(vAppPath) ~ std.path.pathsep ~ GetEnv("PATH"));
    if (vVerbose)
        writefln("Rule Definitions from %s", lRuleDefnFile);
    lRuleText = GetTextLines( lRuleDefnFile);
    foreach(char[] lLine; lRuleText)
    {
        lLine = std.string.strip(lLine);
        if (util.str.begins(lLine, "rule=") > 0 )
        {
            lRules.length = lRules.length + 1;
            lRules[$-1].Name = lLine[5 .. $];
        }
        else if (lRules.length > 0)
        {
            if (util.str.begins(lLine, "in=") > 0)
                lRules[$-1].Input = lLine[3..$];
            else if (util.str.begins(lLine, "out=") > 0)
                lRules[$-1].Output = lLine[4..$];
            else if (util.str.begins(lLine, "tool=") > 0)
                lRules[$-1].Tool = lLine[5..$];
        }
    }

    if (vVerbose)
    {
        foreach(Rule r; lRules)
        {
            writefln("Rule '%s' ==> in:%s, out:%s, tool:'%s'",
                        r.Name, r.Input, r.Output, r.Tool);
        }
    }

    return lRules;
}

//-------------------------------------------------------
int ProcessExternal( ExternRef pRef)
//-------------------------------------------------------
{
    int lResult = -1;
    static Rule[] lRules;
    char[] lExtension;
    char[] lInFile;
    char[] lOutFile;
    FileDateTime lInDate;
    FileDateTime lOutDate;
    char[] lCommand;


    if (lRules.length == 0)
        lRules = LoadRules();

    lExtension = std.path.getExt(pRef.FilePath);
    foreach(Rule r; lRules)
    {
        lInFile = "";
        if (lExtension == r.Input)
        {
            lInFile = pRef.FilePath;
            lOutFile = std.path.addExt(pRef.FilePath, r.Output);

        } else if (lExtension == r.Output)
        {
            lOutFile = pRef.FilePath;
            lInFile = std.path.addExt(pRef.FilePath, r.Input);
        }

        if (lInFile.length == 0)
            continue;

        lInFile = util.pathex.LocateFile(lInFile, vImportRoots);
        if (! util.fileex.FileExists( lInFile ) )
            throw new Exception(
                std.string.format("External file '%s' not found", lInFile));

        lOutFile = getDirName(lInFile) ~ std.path.sep ~ getBaseName(lInFile);
        lOutFile = std.path.addExt(lOutFile, r.Output);

        lInFile = util.pathex.AbbreviateFileName(lInFile);
        lOutFile = util.pathex.AbbreviateFileName(lOutFile);

        lInDate = new FileDateTime(lInFile);
        lOutDate = new FileDateTime(lOutFile);

        if ((lInDate > lOutDate) || vForceCompile)
        {
            char[] lKeyValues;

            lKeyValues = "@IN=" ~ lInFile ~ "," ~ "@OUT=" ~ lOutFile;
            foreach(char[] lOpt; pRef.ToolOpts)
            {
                if (std.string.find(lOpt, "=") != -1)
                    lKeyValues ~= "," ~ lOpt;
            }
            lCommand = util.str.Expand(r.Tool,lKeyValues);

            // Run Compiler to compile the source files that need it.
            lResult = RunCommand(lCommand);
            if (lResult != 0)
                return lResult;
        }
        AddLink(lOutFile);

    }

    return 0;
}

/* Build the target.  Return an error code if there
    was a problem, else return zero.
*/
//-------------------------------------------------------
int Build()
//-------------------------------------------------------
{
    bool        lBuildRequired;
    char[][]    lFilesToLink;
    char[][]    lFilesToCompile;
    int         lRunResult;
    char[]      lTargetName;
    char[]      lTargetDir;
    FileDateTime lTargetTime;
    FileDateTime lMostRecentTime;
    bool        lCompiling;
    bool        lLinking;
    Source[]    lNonLinkingSources;
    char[]      lDResponseFileName;
    char[]      lLinkResponseFileName;
    char[]      lLResponseFileName;
    char[]      lDefName;
    char[]      lOutText;
    char[]      lSourcesToCompile;
    char[]      lCommand;


    // Examine each supplied source file.
    foreach( char[] lFile; vCmdLineSourceFiles)
    {
        new Source(GetFullPathname(lFile));
    }

    // I'm linking if I'm not building a library, and a 'main' was
    // found, and I was not explicitly told not to link.
    lLinking =  (vLibraryAction == LibOpt.Build) ||
                (Source.WasMainFound == false) ||
                (vNoLink == true)
               ? false : true;

    lMostRecentTime = GetNewestDateTime();

    // If not explictly known, set the library action
    // based on whether or not 'main' was found in the sources.
    if (vLibraryAction == LibOpt.Implicit) {
        if (Source.WasMainFound){
            vLibraryAction = LibOpt.DontBuild;
        }
        else {
            vLibraryAction = LibOpt.Build;
        }
    }

    // I'm either creating a library, an executable or neither.
    if (vPragmaTargetName.length != 0)
        vTargetName = vPragmaTargetName;
    if (vCommandTargetName.length != 0)
    {
        char[] lBaseName;

        lBaseName = vTargetName.dup;
        for(int i = lBaseName.length-1; i >= 0; i--)
        {
            if (lBaseName[i] == '.')
            {
                lBaseName.length = i;
                break;
            }
        }
        vTargetName = Expand(vCommandTargetName, "Target=" ~  lBaseName);
    }

    util.pathex.MakePath(vTargetName);

    if (vLibraryAction == LibOpt.Build)
        lTargetName = ReplaceExtention(vTargetName, kLibExtention);
    else if (vNoLink == false)
        if (Source.WasMainFound)
            if (Source.WasMainDLL)
                lTargetName = ReplaceExtention(vTargetName, kShrLibExtention);
            else {
                lTargetName = ReplaceExtention(vTargetName, kExeExtention);
                vTargetExe = lTargetName;
            }
        else
            lTargetName = "not defined";
    else
        lTargetName = "";

    lTargetName = util.pathex.CanonicalPath(lTargetName, false);
    lTargetName = util.pathex.AbbreviateFileName(lTargetName);

    if(vVerbose || vNames) {
        writefln("\nBuilding target '%s'", lTargetName);
    }

    if (lTargetName.length > 0 )
    {
        lTargetName =  util.pathex.CanonicalPath(lTargetName, false);
        lTargetDir = getDirName(lTargetName);

        lTargetTime = util.fdt.GetFileTime (lTargetName);
        if(vVerbose) {
            writefln("Time %s for %s (target)", lTargetTime.toString(), lTargetName);
            writefln("Time %s (most recent)", lMostRecentTime.toString());
        }

        if(lTargetTime < lMostRecentTime) {
            // If the target doesn't exist or it is older than
            // most recently modified dependant file, we need to
            // rebuild the target.
            lBuildRequired = true;
        }
    }

    foreach( ExternRef lRef; vExternals)
    {
        lRunResult = ProcessExternal( lRef );
        if (lRunResult != 0)
            return lRunResult;
    }

    if (Source.SourceIndex.length == 0)
    {
        /* It is possible to only have object and library
           files on the command line, in which case we
           just need to link them rather than compile.
        */

        // No files to compile, just link files, so collect
        // all the object files to link in.
        foreach( char[] lFileName; vLinkFiles.keys)
        {
            // Only include OBJECT files.
            if (util.str.ends(lFileName , kObjExtention) >= 0)
                lFilesToLink ~= lFileName;
        }
    }
    else foreach (int i, char[] lFileName; Source.SourceIndex.keys)
    {
        // Check each source to see if we need to recompile it.
        Source lCurrentSource;
        bool lNeedsCompiling;
        char[] lShortFileName;

        lNeedsCompiling = vForceCompile;

        lCurrentSource = Source.SourceIndex[lFileName];
        if (lCurrentSource.Ignore)
            continue;

        lShortFileName = util.pathex.AbbreviateFileName(lFileName);
        if (std.path.getExt(lShortFileName) != kSrcExtention)
            continue;

        // Only source files are examined from here on.
        if (lCurrentSource.NoLink)
            lNonLinkingSources ~= lCurrentSource;

        if(lCurrentSource.FilesTime > lCurrentSource.ObjectsTime)
        {
            if (vVerbose)
                writefln("%s newer than its object file", lShortFileName);
            lNeedsCompiling=true;

        } else if(lCurrentSource.DependantsTime > lCurrentSource.ObjectsTime) {
            if (vVerbose)
                writefln("%s has newer dependants than its object file.",
                                lShortFileName);

            lNeedsCompiling=true;
        }

        if(lNeedsCompiling)
        {
            lBuildRequired = true;
            lCompiling = true;
            lFilesToCompile ~= lShortFileName;

            // Check to see if I'm allowed to link this file.
            if (lCurrentSource.NoLink == false)
                lFilesToLink ~= util.pathex.AbbreviateFileName(lCurrentSource.ObjectName);

            if (! vTestRun)
            {
                if (lCurrentSource.IncrementBuildNumber())
                {
                    if (vVerbose)
                        writefln("New build number %d for %s",
                               lCurrentSource.BuildNumber,
                               lCurrentSource.ModuleName);

                }
            }
        }
        else if (lCurrentSource.NoLink == false)
        {
            lFilesToLink ~= util.pathex.AbbreviateFileName(lCurrentSource.ObjectName);
        }

    }

    if( lBuildRequired == false )
    {
        if (vSilent == false)
            writefln ("Files are up to date, no build required.");
        return 0;
    }

    lOutText ~= GatherCompilerArgs(lLinking);

    foreach(char[] lFileName; lFilesToCompile)
    {
        lSourcesToCompile ~= util.str.enquote(lFileName) ~ "\n";
    }

    // Construct a Optlink Definition file if requested to.
    version(Windows)
    {
        if (Source.WasMainGUI) {
            AddBuildDef("EXETYPE NT");
            if (vWinVer.length != 0)
                AddBuildDef("SUBSYSTEM WINDOWS," ~ vWinVer);
            else
                AddBuildDef("SUBSYSTEM WINDOWS");
            vDefaultLibs ~= "gdi32.lib";
        }
        else if (Source.WasMainDLL) {
            AddBuildDef(`LIBRARY "` ~ std.path.getBaseName(lTargetName) ~ `"`);
            AddBuildDef("EXETYPE NT");
            if (vWinVer.length != 0)
                AddBuildDef("SUBSYSTEM WINDOWS," ~ vWinVer);
            else
                AddBuildDef("SUBSYSTEM WINDOWS");
            AddBuildDef("CODE PRELOAD DISCARDABLE SHARED EXECUTE");
            AddBuildDef("DATA PRELOAD SINGLE WRITE");
        }
        else if (vConsoleApp) {
            AddBuildDef(`EXETYPE DOS`);
        }

        if ((vNoDef == false) && (vBuildDef.length > 0))
        {
            lDefName = ReplaceExtention(lTargetName, "def");
            if (vTemporaryPath.length != 0)
            {
                lDefName = vTemporaryPath ~ std.path.getBaseName(lDefName);
            }
            lDefName = util.pathex.AbbreviateFileName(lDefName);
            util.fileex.CreateTextFile(lDefName, vBuildDef);

            lOutText ~= util.str.enquote(lDefName) ~ "\n";
        }
    }

    // Add any library or external obj files required.
    if (lLinking) {

        foreach (char[] lFileName; vLinkFiles.keys)
        {
            char[] lCmdItem;
            version(Posix)
                char[] lPrefix;

            if (lFileName.length > 0) {
                lCmdItem = lFileName;
                version(Posix)
                {
                    int lCut;
                    if ( (lCut = util.str.ends(lCmdItem, "." ~ kLibExtention)) >= 0)
                    {
                        lCmdItem = lCmdItem[0 .. lCut]; // Cut off extentsion.
                        lPrefix = "-L-l"; // Needed for linker to find the library.
                    }
                }
                lCmdItem = util.str.enquote(lCmdItem) ~ "\n";

                version(Posix)
                    lOutText ~= lPrefix ~ lCmdItem;
                version(Windows)
                    lOutText ~= lCmdItem;

            }
        }
    }

    if (lLinking || lCompiling)
    {


        // COMPILE phase
        if (lSourcesToCompile.length > 0)
        {
            char[] lCommandLine;

            lCommandLine = lOutText ~ lSourcesToCompile;

            if (vUseResponseFile) {
                lDResponseFileName = ReplaceExtention(lTargetName, "rsp");
                if (vTemporaryPath.length != 0)
                {
                    lDResponseFileName = vTemporaryPath ~ std.path.getBaseName(lDResponseFileName);
                }
                lDResponseFileName = util.pathex.AbbreviateFileName(lDResponseFileName);
                util.fileex.CreateTextFile(lDResponseFileName,lCommandLine);
                lCommand = vCompilerPath ~ vCompiler ~ vCompilerSwitch ~ " @" ~ lDResponseFileName;
            }
            else
            {   // using commandline; may run into limits
                lCommandLine=std.string.replace(lCommandLine, "\n", " ");
                lCommand = vCompilerPath ~ vCompiler  ~ vCompilerSwitch ~ " " ~ lCommandLine;
            }

            if (vVerbose) {
                writefln("Compiling with ..........\n%s\n", lCommandLine);
            }

            // Run Compiler to compile the source files that need it.
            lRunResult = RunCommand(lCommand);
            if (lRunResult != 0)
                vExecuteProgram = false;
        }

        if ( (lRunResult == 0) && (lFilesToLink.length > 0) && (lLinking == true))
        {
            char[] lCommandLine;

            // Ensure the linker gets the right executable name to create.
            char[] lSmallName;
            lSmallName = util.pathex.AbbreviateFileName(lTargetName);
            if (find(lSmallName, " ") != -1) {
                version(Posix) {
                    lOutText ~= std.string.format(`%s"%s"`,vOutFileSwitch,lSmallName);
                }
                else {
                    lOutText ~= std.string.format(`"%s%s"`,vOutFileSwitch,lSmallName);
                }
            }
            else
                lOutText ~= std.string.format(`%s%s`,vOutFileSwitch,lSmallName);

            // Terminate with a newline char.
            lOutText ~= "\n";

            // Include the default libraries.
            foreach( char[] lLib; vDefaultLibs)
            {
                lLib = util.str.enquote(lLib) ~"\n";
                version(Windows)
                    lOutText ~= lLib;
                version(Posix)
                    lOutText ~= "-L-l" ~ lLib;
            }

            if (vLibPaths.length > 1)
            {
                vLibPaths = vLibPaths[1..$];

                // Include the paths to the libraries.
                version(Windows) {

                    if (vVerbose)
                        writefln("Setting LIB=%s", vLibPaths);
                    putenv("LIB=" ~ vLibPaths[1..$]);
                }
                version(Posix)
                {
                    if (vVerbose)
                        writefln("Setting LIB=%s", vLibPaths);
                    char[][] lLibPaths;
                    lLibPaths = std.string.split(vLibPaths, vConfigSep);
                    foreach(char[] lLib; lLibPaths)
                    {
                        if (lLib.length > 0)
                            lOutText ~= "-L-L" ~ lLib ~ "\n";
                    }
                }
            }


            lCommandLine = lOutText;
            foreach(char[] lFile; lFilesToLink)
                lCommandLine ~= lFile ~ "\n";

            if (vUseResponseFile) {
                lLinkResponseFileName = ReplaceExtention(lTargetName, "ksp");
                if (vTemporaryPath.length != 0)
                {
                    lLinkResponseFileName = vTemporaryPath ~ std.path.getBaseName(lLinkResponseFileName);
                }
                lLinkResponseFileName = util.pathex.AbbreviateFileName(lLinkResponseFileName);
                util.fileex.CreateTextFile(lLinkResponseFileName,lCommandLine);
                lCommand = vCompilerPath ~ vLinker ~ " @" ~ lLinkResponseFileName;
            }
            else
            {   // using commandline; may run into limits
                lCommandLine=std.string.replace(lCommandLine, "\n", " ");
                lCommand = vCompilerPath ~ vLinker ~ " " ~ lCommandLine;
            }

            if (vVerbose) {
                writefln("Linking with ..........\n%s\n", lCommandLine);
            }

            // Run Linker
            if (vSilent)
            {
                version(Posix) lCommand ~= " >/dev/null";
                version(Windows) lCommand ~= " >nul";
            }

            lRunResult = RunCommand(lCommand);
            if (lRunResult != 0)
                vExecuteProgram = false;
        }

    }
    else
    {
        if (vSilent == false)
            writefln("No build required.");
        lRunResult = 0;
    }

    // Now build a library if requested to.
    if ( (Source.WasMainDLL)  && (lRunResult == 0) )
    {
        char[] lTargetFileName;
        char[] lImpLibPath;
        ulong[] lImpManf;

        vExecuteProgram = false;

        lImpLibPath = util.pathex.LocateFile("implib.exe", GetEnv("PATH"));
        if (util.fileex.FileExists(lImpLibPath))
        {
            lImpManf = util.fileex.FindInFile(lImpLibPath, "Borland");
            if (lImpManf.length != 0)
                lImpLibPath ~= " -a ";
            else
                lImpLibPath ~= " /system ";

            lTargetFileName = std.path.getBaseName(lTargetName);
            lRunResult = RunCommand(lImpLibPath ~ std.path.addExt(lTargetFileName, "lib")
                            ~ " " ~ std.path.addExt(lTargetFileName, "dll") );
        }
    }
    else if ( (vLibraryAction == LibOpt.Build) && (lRunResult == 0) )
    {
        vExecuteProgram = false;

        lOutText.length = 0;
        lOutText ~= vLibrarianOpts ~ std.path.linesep;
        lOutText ~= lTargetName ~  std.path.linesep;  // Create a new library

        foreach( char[] lFileName; lFilesToLink) {
            char[] lFileDir;

            lFileDir = getDirName(lFileName);
            if (vAllObjects || lFileDir == "" || lFileDir == lTargetDir) {
                if (lFileName.length > 1 + kSrcExtention.length)
                {
                    if (lFileName[$-kSrcExtention.length .. $] == kSrcExtention)
                    {
                        lFileName = lFileName[0..$-kSrcExtention.length] ~ kObjExtention;
                    }
                }

                lOutText ~= lFileName ~ std.path.linesep;
            }
        }

        if (vUseResponseFile) {
            lLResponseFileName = ReplaceExtention(lTargetName, "lsp");
            if (vTemporaryPath.length != 0)
            {
                lLResponseFileName = vTemporaryPath ~ std.path.getBaseName(lLResponseFileName);
            }
            util.fileex.CreateTextFile(lLResponseFileName,lOutText);
            lCommand = vLibrarianPath ~ vLibrarian ~ " @" ~ lLResponseFileName;
        }
        else
        {   // using commandline, may run into limits
            lCommand  = vLibrarianPath ~ vLibrarian ~ " " ~ std.string.replace(lOutText,std.path.linesep," ");
        }

        if (vVerbose) {
            writefln("Librarian with ..........\n%s\n", lOutText);
        }

        lRunResult = RunCommand(lCommand);
    }

    // Optional clean up.
    if (vCleanup) {
        char[][] lHitList;

        if (vVerbose) {
            writefln("Cleaning up ...");
        }
        foreach(char[] lFileName; Source.SourceIndex.keys)
        {
            Source lSource;
            lSource = Source.SourceIndex[lFileName];
            if (lSource.Ignore)
                continue;

            if (lSource.ObjectName.length > 0)
            {
                lHitList ~= lSource.ObjectName;
            }
        }

        // Build's own temprary files.
        lHitList ~= lDResponseFileName;
        lHitList ~= lLinkResponseFileName;
        lHitList ~= lLResponseFileName;
        lHitList ~= lDefName;

        // Possible ones created by compiler, linker, and librarian.
        lHitList ~= ReplaceExtention(lTargetName, "map");
        lHitList ~= ReplaceExtention(lTargetName, "bak");
        lHitList ~= ReplaceExtention(lTargetName, "lst");

        foreach(char[] lFilename; lHitList)
        {
            if (lFilename.length > 0)
            {
                if (util.fileex.FileExists( lFilename ) )
                {
                    if (vVerbose) {
                        writefln("  removing %s", lFilename);
                    }
                    std.file.remove(lFilename);
                }
            }
        }

    }

    return lRunResult;

}

// -------------------------------------------
char [] GatherCompilerArgs(bool pLinking)
// -------------------------------------------
{
    char[] lOutText;

    foreach(char [] lRoot; vImportRoots)
    {
        version(Posix)
            vDefaultCompArgs ~= vImportPathSwitch ~ "\"" ~ lRoot ~ "\"";
        else
            vDefaultCompArgs ~= vImportPathSwitch ~ lRoot;
    }

    foreach(char[] lArg; vDefaultCompArgs)
    {
        AddCompilerArg( lArg );
    }

    // Build command files for compilation.
    foreach (char[] lCompileArg; vCompilerArgs)
    {
        // Ignore empty args
        if (lCompileArg.length > 0)
        {
            // Enclose in quotes if no quotes are currently present.
            if (find(lCompileArg, "\"") == -1)
                // Arguments containing a blank need to be quoted.
                if (find(lCompileArg, " ") != -1)
                {
                    // Strip off any trailing shell escape lead-in character.
                    if (lCompileArg[$-1] == '\\')
                        lCompileArg.length = lCompileArg.length - 1;
                    lOutText ~= std.string.format(`"%s"`, lCompileArg);
                }
                else
                    lOutText ~= lCompileArg;
            else
                lOutText ~= lCompileArg;

            // Terminate with a newline char.
            lOutText ~= "\n";
        }
    }

    if (pLinking == false)
    { // No linking allowed.
        if (vNoLinkSwitch.length > 0)
        {
            if (find(vNoLinkSwitch, " ") != -1)
                lOutText ~= std.string.format(`"%s"`,vNoLinkSwitch);
            else
                lOutText ~= vNoLinkSwitch;
            // Terminate with a newline char.
            lOutText ~= "\n";

            AddCompilerArg( vNoLinkSwitch );
        }
    }

    return lOutText;
}


// -------------------------------------------
char[][] ModulesToIgnore()
// -------------------------------------------
{
    return vModulesToIgnore;
}


// -------------------------------------------
void AddExternal(char[] pPath, char[][] pOpts)
// -------------------------------------------
{
    if (pPath.length > 0) {
        vExternals.length = vExternals.length + 1;
        vExternals[$-1].FilePath = pPath;
        vExternals[$-1].ToolOpts = pOpts;
        if (vVerbose) {
            writef("Added external file to be built: %s", pPath);
            foreach( char[] lOpt; pOpts)
                writef(" `%s`", lOpt);
            writefln("");
        }
    }
}

// -------------------------------------------
void AddLink(char[] pPath)
// -------------------------------------------
{
    if (pPath.length == 0)
        return;
    if ((pPath in vLinkFiles) is null)
        vLinkFiles[pPath] = true;
}

// -------------------------------------------
void AddTarget(char[] pPath)
// -------------------------------------------
{
    if (pPath.length == 0)
        return;

    if (vPragmaTargetName.length == 0)
        vPragmaTargetName = pPath;
    else if (vPragmaTargetName != pPath) {
        if (vVerbose) {
            writefln("Multiple pragma(target,...) detected. '%s' will be used and '%s' rejected.",
                    vPragmaTargetName, pPath );
        }
    }

}

// -------------------------------------------
void AddBuildDef(char[] pText, bool pReplace = false)
// -------------------------------------------
{
    int lPos;
    char[] lLowerText;
    static uint[ char[] ] lElementIdx;

    if (vNoDef == true)
        return;

    lLowerText = std.string.tolower(pText);

    lPos = std.string.find(lLowerText, ' ');
    if (lPos == -1)
        lPos = lLowerText.length;
    lLowerText.length = lPos;

    if (lLowerText in lElementIdx)
    {
        if (pReplace)
            vBuildDef[ lElementIdx[lLowerText] ] = pText;
    }
    else {
        vBuildDef ~= pText;
        lElementIdx[ lLowerText ] = vBuildDef.length-1;
    }
}

// -------------------------------------------
void AddCompilerArg(char[] pArg)
// -------------------------------------------
{
    bool lFound;

    if (pArg.length > 0)
    {
        // Translate exported version pragmas.
        if (pArg.length > 3)
        {
            if (pArg[0..3] == `+v+`)
                pArg = vVersionSwitch ~ pArg[3..$];
        }

        lFound = false;
        foreach(char[] lArg; vCompilerArgs)
        {
            if (lArg == pArg)
            {
                lFound = true;
                break;
            }
        }
        if (! lFound )
            vCompilerArgs ~= pArg;
    }
}

// -------------------------------------------
char[] AddRoot(char[] pRootName)
// -------------------------------------------
{
    static bool [char[]] lRootHash;
    char[] lFullName;
    char[] lSearchName;

    if(pRootName.length == 0) {
        return pRootName;
    }
    else {
    lFullName = util.pathex.CanonicalPath(pRootName);
    version(Windows) lSearchName = std.string.tolower(lFullName);
    version(Posix)   lSearchName = lFullName;
    if( !(lSearchName in lRootHash) ) {
        vImportRoots ~= lFullName;
        lRootHash[lSearchName] = true;
        return lFullName;
    }
    else
        return "";
}
}

bool AutoImports()
{
    return vAutoImports;
}

char[][] GetImportRoots()
{
    return vImportRoots;
}

// function to replace tokens in the form %<SYM>%  with environment data.
// -------------------------------------------
char[] ExpandEnvVar(char[] pLine)
// -------------------------------------------
{
    char[] lLine;
    char[] lSymName;
    int lPos;
    int lEnd;

    lLine.length = 0;
    for( lPos = 0; lPos < pLine.length; lPos++ )
    {
        if (pLine[lPos] == '%')
        {
            lSymName.length = 0;
            for(lEnd = lPos+1; (lEnd < pLine.length) && (pLine[lEnd] != '%'); lEnd++ )
            {
                lSymName ~= pLine[lEnd];
            }
            if (lSymName.length > 0)
            {
                if (lSymName == "@P")
                {
                    // Special symbol that refers to the DMD configuration file's directory.
                    lSymName = (vConfigPath ? vConfigPath[0..length-1] : "");
                }
                else if (lSymName == "@D")
                {
                    // Special symbol that refers to the compiler's file's directory.
                    lSymName = (vCompilerPath ? vCompilerPath[0..length-1] : "");
                }
                else
                {
                    lSymName = GetEnv(lSymName);
                }
                lLine ~= lSymName;
            }
            lPos = lEnd;
        }
        else
        {
            lLine ~= pLine[lPos];
        }
    }
    return lLine;
}

void Process_DFLAGS(char[] pText)
{
    int      lPos;
    int      lEndPos;
    char[]   lRootName;
    char[][] lRoots;
    char[][] lArgs;
    int      lArg;
    bool     lInArg;
    char     lQuote;

    lInArg = false;
    lArg = -1;
    lQuote = 0;
    foreach (char lArgChar; pText)
    {
        if ( (lArgChar == '"') || (lArgChar == '\'') )
        {
            if (lQuote == lArgChar)
            {
                lQuote = 0;
                continue;
            }

            if (lQuote == 0)
            {
                lQuote = lArgChar;
                continue;
            }
        }

        if (lArgChar == ' ')
        {
            if (lQuote == 0)
            {
                lInArg = false;
                continue;
            }
        }

        if (lInArg == false)
        {
            lArg++;
            lArgs.length = lArg+1;
            lInArg = true;
        }
        lArgs[lArg] ~= lArgChar;

    }

    foreach(char[] lSwitch; lArgs)
    {
        if ((lSwitch.length > 0) && (lSwitch[0] == '-'))
        {
            if (vDelayedValue != null)
            {
                // Used when an switch needs the subsequent arg to
                // be its value.
                *vDelayedValue = lSwitch;
                vDelayedValue = null;
                vBuildArgs ~= lSwitch;
                continue;
            }

            if (lSwitch[1] == 'I')
            {
                lRoots = std.string.split(lSwitch[2..length], vConfigSep);
                foreach(char[] lRoot; lRoots)
                {
                    lRootName = AddRoot(lRoot);
                    /*
                    version(Posix)
                        vDefaultCompArgs ~= vImportPathSwitch ~ "\"" ~ lRoot ~ "\"";
                    else
                        vDefaultCompArgs ~= vImportPathSwitch ~ lRoot;
                        */
                    if(vVerbose) {
                        if (lRootName.length > 0){
                            writefln(" added root from config file %s", lRootName);
                        }
                    }
                }
            }
            else
            {
                version(DigitalMars)
                {
                if (util.str.IsLike(lSwitch,  (vOutFileSwitch ~ "*")))
                {
                    // Target name (eg. -oftestapp)
                    vCommandTargetName = lSwitch[vOutFileSwitch.length .. $];
                    vBuildArgs ~= lSwitch;
                    continue;
                }
                }

                version(GNU)
                {
                    if (lSwitch == vOutFileSwitch)
                    {
                        // Target name (eg. -o testapp)
                        vDelayedValue = &vCommandTargetName;
                        vBuildArgs ~= lSwitch;
                        continue;
                    }
                }

                if (util.str.IsLike(lSwitch,  std.utf.toUTF32(vTemporaryPathSwitch ~ "*")))
                {
                    char[] lbRoot;

                    vTemporaryPath = lSwitch[vTemporaryPathSwitch.length .. $];
                    if (vTemporaryPath.length > 0 && vTemporaryPath[$-1..$] != std.path.sep)
                        vTemporaryPath ~= std.path.sep;

                    version(DigitalMars)
                    {
                        lbRoot = AddRoot(vTemporaryPath);
                        if (lbRoot.length > 0){
                            if(vVerbose) {
                                    writefln("Added root from config file Object Write Path = %s",lbRoot);
                            }
                            util.pathex.MakePath(lbRoot);
                        }
                    }

                    vBuildArgs ~= lSwitch;
                    // This one actually *is* passed thru.
                }

                AddCompilerArg ( lSwitch );
            }
        }
    }
}

// -------------------------------------------
void ReadEnviron()
// -------------------------------------------
{
    char[]   lSymValue;

    // Check for a environment flag before config file.
    lSymValue = GetEnv("DFLAGS");
    if (lSymValue.length > 0 )
    {
        if (vVerbose)
            writefln("Analyzing environment symbol DFLAGS=%s", lSymValue);
        Process_DFLAGS( lSymValue );
    }
}

// -------------------------------------------
void ReadConfigFile()
// -------------------------------------------
{
    char[]   lPath;
    char[][] lTextLines;
    int      lPos;

    if (vConfigFile is null)
    {
        // There is no configuration file to process
      return;
    }

    lPath = vConfigPath ~ vConfigFile ;
    lTextLines = util.fileex.GetTextLines(lPath, util.fileex.GetOpt.Exists);

    if (vVerbose) {
        writefln("Reading from config: %s", lPath);
    }

    foreach(int i, char[] lLine; lTextLines)
    {
        // Strip off trailing whitespace.
        while(lLine.length > 0 && lLine[lLine.length-1] <= ' ') {
            lLine.length = lLine.length-1;
        }
        // Replace any environment symbols with their value.
        if (vVerbose) {
            writefln(" Line %d: %s", i+1, lLine);
        }
        lLine = ExpandEnvVar(lLine);

        // Examine DFLAGS
        lPos = find(lLine, "DFLAGS=");
        if(lPos == 0)
        {
            Process_DFLAGS(lLine[lPos+7..length]);
        } // end of DFLAGS processing.

        // Examine LIB
        lPos = find(lLine, "LIB=");
        if(lPos == 0)
        {
            char[][] lPaths;
            lLine = lLine[lPos+4 .. length];
            lPaths.length = 1;
            foreach(char lChar; lLine)
            {
                if (lChar == '"') {}
                else if (lChar == ';')
                {
                    lPaths.length = lPaths.length + 1;
                }
                else
                    lPaths[$-1] ~= lChar;

            }
            for(int i = 0; i < lPaths.length; i++)
            {
                lPaths[i] = std.string.strip(lPaths[i]);
            }

            foreach( char[] lPath; lPaths)
            {
                if (lPath.length > 0)
                {
                    if (lPath[0] == '"' && lPath[length-1] == '"') {
                        lPath = lPath[1..length-1];
                    }
                    vLibPaths ~= vConfigSep;
                    vLibPaths ~= ("\"" ~ util.pathex.CanonicalPath(lPath) ~ "\"");
                }
            }
            if(vVerbose)
            {
                writefln(" use %s",vLibPaths);
            }
            continue;
        }

        // Examine LINKCMD
        lPos = find(lLine, "LINKCMD=");
        if(lPos == 0) {
            // Strip out any quotes
            while( (lPos = find(lLine, "\"")) != -1)
            {
                lLine = lLine[0..lPos] ~ lLine[lPos+1 .. $];
            }

            if (vLibrarianPath.length == 0)
            {
                vLibrarianPath = std.path.getDirName(lLine[8..$]) ~ std.path.sep;
                vLibrarianPath = util.pathex.CanonicalPath(vLibrarianPath);
                if(vVerbose) {
                    writefln(" librarian path %s",vLibrarianPath);
                }
            }
            continue;
        }

        lPos = find(lLine, "LIBCMD=");
        if(lPos == 0) {
            // Strip out any quotes
            while( (lPos = find(lLine, "\"")) != -1)
            {
                lLine = lLine[0..lPos] ~ lLine[lPos+1 .. $];
            }
            vLibrarianPath = std.path.getDirName(lLine[7..$]) ~ std.path.sep;
            vLibrarianPath = util.pathex.CanonicalPath(vLibrarianPath);

            vLibrarian = std.path.getBaseName(lLine[7..$]);
            if(vVerbose) {
                writefln(" librarian path %s",vLibrarianPath);
                writefln(" librarian is %s",vLibrarian);
            }
        }
    }
}

// Display each entry in the supplied list.
// -------------------------------------------
void DisplayItems(char[][] pList, char[] pTitle = "")
// -------------------------------------------
{
    if (pList.length > 0) {
        if (pTitle.length > 0)
            writefln("\n%s",pTitle);

        foreach(int lIndex, char[] lListEntry; pList) {
            writefln(" [%2d]: %s",lIndex,lListEntry);
        }
    }
}

// -------------------------------------------
void DisplayItems(ExternRef[] pList, char[] pTitle = "")
// -------------------------------------------
{
    if (pList.length > 0) {
        if (pTitle.length > 0)
            writefln("\n%s",pTitle);

        foreach(int lIndex, ExternRef lListEntry; pList)
        {
            writef(" [%2d]: %s",lIndex,lListEntry.FilePath);
            if (lListEntry.ToolOpts.length > 0)
                foreach(char[] lOpt; lListEntry.ToolOpts)
                    writef(" [%s]",lOpt);
            writefln("");
        }
    }
}

// ------------------------------------------------
char[] GetTemporaryPath()
{
    return vTemporaryPath;
}

// ------------------------------------------------
Substitute[] GetMacros()
{
    return vSubstitutions;
}

// ------------------------------------------------
char[] GetFullPathname(char[] pFileName)
// -------------------------------------------
{
    char[] lPossiblePath;
    char[] lLocalPath;

    // Look for file in current folder first.
    lLocalPath = util.pathex.CanonicalPath(pFileName, false);
    if(util.fileex.FileExists(lLocalPath)) {
        return util.pathex.AbbreviateFileName(lLocalPath);
    }

    // Examine each known import root to see if the file lives there.
    foreach(char[] lNextRoot; vImportRoots) {
        lPossiblePath = ( lNextRoot ~ pFileName );
        if(util.fileex.FileExists(lPossiblePath)) {
            return util.pathex.AbbreviateFileName(lPossiblePath);
        }
    }

    return util.pathex.AbbreviateFileName(lLocalPath);
}


// Function to locate where Complier is installed from the PATH symbol
// -------------------------------------------
char[] FindFileInPathList(char[] pSymName, char[] pFileName)
// -------------------------------------------
{
    char[][] lPaths;
    char[]   lCompilerPath;
    char[]   lRawValue;

    // Assume that an environment symbol name was supplied,
    // but if that fails, assume its a list of paths.
    lRawValue = GetEnv(pSymName);
    if (lRawValue.length == 0)
        lRawValue = pSymName;

    // Rearrange path list into an array of paths.
    lPaths = split(util.str.toASCII(lRawValue), std.path.pathsep);

    lCompilerPath.length = 0;
    foreach(char[] lPath; lPaths)
    {
        if (lPath.length > 0)
        {
            // Ensure that the path ends with a valid separator.
            if (lPath[length-1] != std.path.sep[0] )
                lPath ~= std.path.sep;
            // If the file is in the current path we can stop looking.
            if(util.fileex.FileExists(lPath ~ pFileName))
            {
                // Return the path we actually found it in.
                lCompilerPath = lPath;
                break;
            }
        }
    }

    return lCompilerPath;
}

// -------------------------------------------
int main(char[][] pArgs)
// -------------------------------------------
{
    int lBuildResult;
    char[] lCompPath;
    bool lSetPath = false;

    /* Set routine addresses in the Source module. This allows
       the methods and functions in Source to access the optional
       functionality provided by this module.
    */
    source.AddRoot          = &AddRoot;
    source.GetImportRoots   = &GetImportRoots;
    source.ModulesToIgnore  = &ModulesToIgnore;
    source.AutoImports      = &AutoImports;
    source.AddTarget        = &AddTarget;
    source.AddLink          = &AddLink;
    source.AddExternal      = &AddExternal;
    source.AddBuildDef      = &AddBuildDef;
    source.GetFullPathname  = &GetFullPathname;
    source.GetObjWritePath  = &GetTemporaryPath;
    source.GetMacros        = &GetMacros;
    source.AddCompilerArg   = &AddCompilerArg;

    // Scan the PATH env symbol to locate the D compiler.
    lCompPath = FindFileInPathList("PATH", vCompiler);
    if (lCompPath.length > 0){
        if (lCompPath[length-1] != std.path.sep[0])
            lCompPath ~= std.path.sep;
        vCompilerPath = lCompPath.dup;
        // Set config path to same as compiler unless it is already set.
        if (!(vConfigPath is null) && (vConfigPath.length == 0))
            vConfigPath = lCompPath.dup;
    }

    if (vCompilerPath.length == 0)
    {
        vCompilerPath = util.pathex.GetInitCurDir;
        lSetPath = true;
    }

    // Strip off application's path from arglist.
    vAppPath=pArgs[0];
    vAppName=getBaseName(vAppPath);
    version (Windows)
    {
        int DotPos;
        DotPos = rfind(vAppName, '.');
        if (DotPos != -1)
        {
            vAppName.length = DotPos;
        }
    }

    pArgs=pArgs[1..pArgs.length];
    if (pArgs.length == 0) {
        // No other arguments so show usage message.
        DisplayUsage ();
        return 0;
    }


    GatherArgs( pArgs );
    source.mVerboseMode = vVerbose;
    if (vVerbose){
        writefln("*** build v%s (build %d)***", vAppVersion, build_bn.auto_build_number);
    }

    foreach(char[] lArg; vCombinedArgs) {
        ProcessCmdLineArg( lArg );
    }

    source.mMacroInput = vMacroInput;
    if (vMacroInput)
        ProcessMacroDefs(vVerbose);

    if( (vTargetName.length == 0) && (vCommandTargetName.length == 0) ){
        throw new Exception("No target name supplied.");
    }

    if (vVerbose || vNames)
        writefln("Current Dir '%s'", util.pathex.GetInitCurDir());

    if (lSetPath)
    {
        if (vVerbose)
        {
            writefln("%s not found in PATH symbol, so assuming current directory",
                        vCompiler);
        }
    }

    if (vCompilerPath[length-1] != std.path.sep[0])
        vCompilerPath ~= std.path.sep;


    if (vConfigPath.length == 0)
    {
        vConfigPath = vCompilerPath;
        if (vVerbose)
        {
            writefln("No configuration file path, so assuming current directory");
        }
    }
    if (vConfigPath[length-1] != std.path.sep[0])
        vConfigPath ~= std.path.sep;

    version(DigitalMars)
    {
        if (util.fileex.FileExists(vConfigPath ~ vConfigFile) == false)
        {
            throw new Exception(std.string.format("The configuration file '%s' was not found.",
                    vConfigPath ~ vConfigFile));
        }
    }

    if (util.fileex.FileExists(vCompilerPath ~ vCompiler) == false)
    {
        throw new Exception(std.string.format("The compiler '%s' was not found.",
                vCompilerPath ~ vCompiler));
    }

    if (vVerbose){
        writefln("Compiler installed in %s",vCompilerPath);
        writefln("Configuration File installed in %s",vConfigPath);
    }

    source.SetKnownVersions();

    ReadEnviron();
    version(DigitalMars) ReadConfigFile();

    // Rationalize the ignored modules list.
    foreach(char[] m; vModulesToNotice) {
        for (int i=0; i < vModulesToIgnore.length; i++) {
            if (vModulesToIgnore[i] == m) {
                // Must remove from ignored list.
                vModulesToIgnore = vModulesToIgnore[0..i] ~ vModulesToIgnore[i+1..length];
                i--;
            }
        }
    }


    // Process the D Source files that were on the command line.
    // Each file must already exist. If any does not exist, then
    // the application will abort.
    bool lAllExist = true;

    // List all missing files (if any).
    foreach( char[] lFile; vCmdLineSourceFiles)
    {
        if (! util.fileex.FileExists( lFile ) )
        {
            if (vSilent == false)
                writefln("** File '%s' not found.", lFile);
            lAllExist = false;
        }
    }

    if ( lAllExist == false)
    {
        throw new Exception ("Not all supplied files exist.");
    }

    lBuildResult = Build();

    if(vVerbose) {
        writefln("");
        DisplayItems(vBuildArgs,      "build args: ...............");
        DisplayItems(vCompilerArgs,   "compiler args: ................");
        DisplayItems(vCmdLineSourceFiles,   "command line files: ...............");
        DisplayItems(Source.SourceIndex.keys,
                                      "source files: ...............");
        DisplayItems(vLinkFiles.keys, "link files: ...............");
        DisplayItems(vExternals,      "externally built files: ...............");
        DisplayItems(vImportRoots,    "import roots: .................");
        DisplayItems(vModulesToIgnore,"ignored packages: .................");
        DisplayItems(vModulesToNotice,"noticed package: .................");
    }

    if (vExecuteProgram && (vTargetExe.length > 0))
    {
        RunCommand( vTargetExe ~ " " ~ std.string.strip(vRunParms));
    }

    return lBuildResult;


}

void ProcessResponseFile(char[] pArg, bool pVerbose)
{
    // A response file is being used.
    char[] lRespFileName;
    char[][] lRespLines;

    if (pArg.length == 1)
        lRespFileName = "@build.brf";
    else
        lRespFileName = pArg[1..length].dup;

    if (std.path.getExt(lRespFileName).length == 0)
    {
        lRespFileName ~= ".brf";
    }

    if (pVerbose)
        writefln("Response file %s", lRespFileName);

    lRespLines = util.fileex.GetTextLines(lRespFileName, util.fileex.GetOpt.Exists);
    foreach(char[] lArg; lRespLines)
    {
        // Locate any comment text in the line.
        int lPos = std.string.find(lArg, "#");
        if (lPos != -1)
        {
            // Truncate the line at the '#' character.
            lArg.length = lPos;
        }

        lArg = std.string.strip(lArg);
        if (lArg.length > 1){
            if (pVerbose)
                writefln("Response file arg: %s", lArg);
            GatherOneArg( lArg );
        }
    }
}

void ProcessCmdLineArg( char[] pArg )
{
    char[] lNewPath;
    static char[] lImportSwitch;
    int lCut;

    if(util.str.begins(pArg, "-version=") > 0) {
        char[] lVersionString;

        lVersionString=pArg [9 .. $];
        source.ActivateVersion(lVersionString);
        pArg = vVersionSwitch ~ lVersionString;
    }
    else if(util.str.begins(pArg, "-fversion=") > 0) {
        char[] lVersionString;

        lVersionString=pArg [10 .. $];
        source.ActivateVersion(lVersionString);
        pArg = vVersionSwitch ~ lVersionString;
    }

    else if(pArg == "-fdebug") {
        source.ActivateDebug("1");
        pArg = vDebugSwitch;
    }
    else if(pArg == "-debug") {
        source.ActivateDebug("1");
        pArg = vDebugSwitch;
    }
    else if(util.str.begins(pArg, "-fdebug=") > 0) {
        char[] lDebugString;

        lDebugString=pArg [8 .. $];
        source.ActivateDebug(lDebugString);
        pArg = vDebugSwitch ~ "=" ~ lDebugString;
    }
    else if(util.str.begins(pArg, "-debug=") > 0) {
        char[] lDebugString;

        lDebugString=pArg [7 .. $];
        source.ActivateDebug(lDebugString);
        pArg = vDebugSwitch ~ "=" ~ lDebugString;
    }

    switch(pArg) {
        case "-full":
            vForceCompile = true;
            vBuildArgs ~= pArg;
            // Not passed thru.
            break;

        case "-link":
            Source.WasMainFound = true;
            vBuildArgs ~= pArg;
            // Not passed thru.
            break;

        case "-nolink":
            vNoLink = true;
            vBuildArgs ~= pArg;
            // Not passed thru.
            break;

        case "-lib":
            vLibraryAction = LibOpt.Build;
            vBuildArgs ~= pArg;
            // Not passed thru.
            break;

        case "-nolib":
            vLibraryAction = LibOpt.DontBuild;
            vBuildArgs ~= pArg;
            // Not passed thru.
            break;

        case "-obj":
            vLibraryAction = LibOpt.DontBuild;
            vNoLink = true;
            vBuildArgs ~= pArg;
            // Not passed thru.
            break;

        case "-nounittest":
            // Not passed thru. Deprecated switch is now ignored.
            if (vSilent == false)
                writefln("Note: '-nounittest' ignored. This switch is no longer used.");
            break;

        case "-info":
            DisplayUsage(false);
            vBuildArgs ~= pArg;
            // Not passed thru.
            break;

        case "-silent":
            vSilent = true;
            vBuildArgs ~= pArg;
            // Not passed thru.
            break;

        case "-noautoimport":
            vAutoImports = false;
            vBuildArgs ~= pArg;
            // Not passed thru
            break;

        case "-nodef":
            vNoDef = true;
            vBuildArgs ~= pArg;
            // Not passed thru.
            break;

        case "-nomacro":
            vMacroInput = false;
            vBuildArgs ~= pArg;
            // Not passed thru.
            break;

        case "-usage":
        case "-help":
        case "-h":
        case "-?":
            DisplayUsage();
            vBuildArgs ~= pArg;
            // Not passed thru.
            break;

        case "-allobj":
            vAllObjects = true;
            vBuildArgs ~= pArg;
            // Not passed thru.
            break;

        case "-test":
            vTestRun = true;
            vBuildArgs ~= pArg;
            // Not passed thru.
            break;

        case "-cleanup":
            // drop through ...
        case "-clean":
            vCleanup = true;
            vBuildArgs ~= pArg;
            // Not passed thru.
            break;

        case "-V": /* we need verbose status earlier */
            vBuildArgs ~= pArg;
            // Not passed thru.
            break;

        case "-names":
            vBuildArgs ~= pArg;
            vNames = true;
            // Not passed thru.
            break;

        default:
            if (pArg.length > 0)
            {
            if (pArg[0] == '-') {
                // Test for Librarian options.
                if (util.str.IsLike(pArg, cast(dchar[])"-LIBOPT*"))
                {
                    vLibrarianOpts ~= " " ~ pArg[7..$].dup;
                    vBuildArgs ~= pArg;
                    break;

                }

                if (util.str.begins(pArg, "-LIBPATH=") >= 0)
                {
                    vLibPaths ~= vConfigSep ~ pArg[9..$].dup;
                    vBuildArgs ~= pArg;
                    break;
                }

                // Test for alternate install locations.
                if (util.str.IsLike(pArg, cast(dchar[])"-DCPATH*"))
                {
                    lNewPath = pArg[7..length].dup;
                    if (vCompilerPath == vConfigPath)
                    {
                        if (vVerbose)
                            writefln("CFPATH was %s now %s", vConfigPath, lNewPath);
                        vConfigPath = lNewPath;
                    }
                    if (vVerbose)
                        writefln("DCPATH was %s now %s", vCompilerPath, lNewPath);
                    vCompilerPath = lNewPath;
                    vBuildArgs ~= pArg;
                    break;

                }

                if (util.str.IsLike(pArg,  cast(dchar[])"-CFPATH*"))
                {
                    if (vVerbose)
                        writefln("CFPATH was %s now %s", vConfigPath, pArg[7..length]);
                    vConfigPath = pArg[7..length].dup;
                    vBuildArgs ~= pArg;
                    break;

                }

                if (util.str.IsLike(pArg,  cast(dchar[])"-RDF*"))
                {
                    if (vVerbose)
                        writefln("RDF was %s now %s", vRDFName, pArg[4..$]);
                    vRDFName = pArg[4..$].dup;
                    vBuildArgs ~= pArg;
                    break;

                }

                if (pArg == vBuildImportPathSwitch)
                {
                    vDelayedValue = &lImportSwitch;
                    vBuildArgs ~= pArg;
                    break;
                }
                else if ( (lCut = util.str.begins(pArg, vBuildImportPathSwitch)) >= 0)
                {
                    char [] lRoot;
                    foreach(char[] lCmdRoot; std.string.split(pArg[lCut..$], ";"))
                    {
                        lRoot = AddRoot(lCmdRoot);
                        if (lRoot.length > 0){
                            if(vVerbose) {
                                    writefln("Added root from command line = %s",lRoot);
                            }
                        }
                    }
                    break;
                }
                if (util.str.IsLike(pArg,  std.utf.toUTF32(vRunSwitch ~ "*")))
                {
                    vRunParms ~= pArg[vRunSwitch.length .. $] ~ " ";
                    vExecuteProgram = true;
                    vBuildArgs ~= pArg;
                    break;
                }


                // Special check for Object Write Path
                version(DigitalMars)
                {
                if (util.str.IsLike(pArg,  std.utf.toUTF32(vTemporaryPathSwitch ~ "*")))
                {
                    char[] lRoot;

                    vTemporaryPath = pArg[vTemporaryPathSwitch.length .. $];
                    if (vTemporaryPath.length > 0 && vTemporaryPath[$-1..$] != std.path.sep)
                        vTemporaryPath ~= std.path.sep;

                    lRoot = AddRoot(vTemporaryPath);
                    if (lRoot.length > 0){
                        if(vVerbose) {
                                writefln("Added root from Object Write Path = %s",lRoot);
                        }
                        util.pathex.MakePath(lRoot);
                    }

                    vBuildArgs ~= pArg;
                    // This one actually *is* passed thru.
                }
                }

                if (util.str.IsLike(pArg,  cast(dchar[])"-X*"))
                {
                    // Modules to ignore (eg. -Xmylib)
                    vModulesToIgnore ~= pArg[2..$];
                    vBuildArgs ~= pArg;
                    break;
                }

                if (util.str.IsLike(pArg,  cast(dchar[])"-M*"))
                {
                    // Modules to notice (eg. -Mphobos)
                    vModulesToNotice ~= pArg[2..$];
                    vBuildArgs ~= pArg;
                    break;
                }

                if (util.str.IsLike(pArg,  cast(dchar[])"-T*"))
                {
                    // Target name (eg. -Ttestapp)
                    vCommandTargetName = pArg[2..$];
                    vBuildArgs ~= pArg;
                    break;
                }

                version(DigitalMars)
                {
                if (util.str.IsLike(pArg,  (vOutFileSwitch ~ "*")))
                {
                    // Target name (eg. -oftestapp)
                    vCommandTargetName = pArg[vOutFileSwitch.length .. $];
                    vBuildArgs ~= pArg;
                    break;
                }
                }

                version(GNU)
                {
                if (pArg == vOutFileSwitch)
                {
                    // Target name (eg. -o testapp)
                    vDelayedValue = &vCommandTargetName;
                    vBuildArgs ~= pArg;
                    break;
                }
                }

                if (util.str.IsLike(pArg,  cast(dchar[])"-R*"))
                {
                    // Response file usage (eg. -Ry)
                    if (pArg.length == 2)
                        vUseResponseFile = (vUseResponseFile ? false : true);

                    else if (pArg[2] == 'y')
                        vUseResponseFile = true;

                    else
                        vUseResponseFile = false;

                    vBuildArgs ~= pArg;
                    break;
                }

                version(Windows) {
                    if (util.str.IsLike(pArg,  cast(dchar[])"-gui*"))
                    {
                            Source.WasMainGUI = true;
                            vBuildArgs ~= pArg;
                            if (pArg.length == 4)
                                break;
                            if (pArg[4] == ':')
                                vWinVer = pArg[5..length];
                            else
                                vWinVer = pArg[4..length];
                            break;
                    }
                    if (pArg == "-dll")
                    {
                            Source.WasMainDLL = true;
                            vBuildArgs ~= pArg;
                            break;
                    }
                }

                AddCompilerArg( pArg );

            } else {

                if (vDelayedValue != null)
                {
                    // Used when an switch needs the subsequent arg to
                    // be its value.
                    if ( vDelayedValue == &lImportSwitch)
                    {
                        char [] lRoot;
                        foreach(char[] lCmdRoot; std.string.split(pArg, ";"))
                        {
                            lRoot = AddRoot(lCmdRoot);
                            if (lRoot.length > 0){
                                if(vVerbose) {
                                        writefln("Added root from command line = %s",lRoot);
                                }
                            }
                        }
                    }
                    else
                    {
                        *vDelayedValue = pArg;
                        vBuildArgs ~= pArg;
                    }
                    vDelayedValue = null;
                    break;
                }

                switch(std.path.getExt(pArg)) {
                   case "":
                       pArg ~= "." ~ kSrcExtention;
                       vCmdLineSourceFiles ~= pArg;
                       break;

                   case kSrcExtention:
                   case kMacroExtention:
                       vCmdLineSourceFiles ~= pArg;
                       break;

                   default:
                       AddLink(GetFullPathname(pArg));
                }
                if(vTargetName is null) {
                   vTargetName = pArg;
                }
            }
            break;
        }
    }
}

void GatherOneArg( char[] pArg )
{
    static bool[char[]] lKnownArgs;

    pArg = std.string.strip(pArg);
    if (pArg.length == 0)
        return;

    if ((pArg == "-V") || (pArg == "-v"))
        vVerbose = true;

    if ( pArg.length >= 2 && pArg[0..2] == "--")
    {   // Need to remove an earlier matching argument.
        pArg = pArg[1..$];
        if ((pArg == "-V") || (pArg == "-v"))
            vVerbose = false;
        for(int i = vCombinedArgs.length-1; i >= 0; i--)
        {
            if (pArg[$-1] == '*')
            {
                if (vCombinedArgs[i].begins(pArg[0..$-1]) > 0)
                {
                    vCombinedArgs = vCombinedArgs[0..i] ~
                                vCombinedArgs[i+1 .. $];
                    break;
                }
            }
            else if (vCombinedArgs[i] == pArg)
            {
                vCombinedArgs = vCombinedArgs[0..i] ~
                                vCombinedArgs[i+1 .. $];
                break;
            }
        }
    }
    else if (pArg[0] == '@')
        ProcessResponseFile(pArg, vVerbose);
    else if (pArg[0] == '+')
        ProcessBuildConfig(pArg, vVerbose);
    else
    {   // Only add an argument if it is not already been added.
        if ( !(pArg in lKnownArgs) )
        {
            vCombinedArgs ~= pArg;
            lKnownArgs[pArg] = true;
        }
    }
}

void GatherArgs( char[][] pArgs )
{
    /* This collects together all the original command line arguments,
       any command line arguments in any configuration files, and
       the contents of any response files.
    */

    // Collect from configuration file(s).
    ProcessBuildConfig("+", vVerbose);

    // Collect from original command line.
    foreach( char[] lArg; pArgs)
    {
        GatherOneArg( lArg );
    }


}

void ProcessMacroDefs(bool pVerbose)
{
    // From build.exe location
    ProcessOneMacroDef(pVerbose, std.path.getDirName(vAppPath));

    // From compiler location
    ProcessOneMacroDef(pVerbose, vCompilerPath);

}

void ProcessOneMacroDef(bool pVerbose, char[] pPath)
{
    char[] lMacroDefFileName;
    char[][] lMacroDefLines;
    char[] lOpen;
    char[] lClose;
    char[] lEscapeOpen;
    char[] lEscapeClose;
    char[] lCommentLead;
    static char[] lDefOpen = "\"";
    static char[] lDefClose = "\"";
    static char[] lDefEscapeOpen = "\\";
    static char[] lDefEscapeClose = "";
    static char[] lDefCommentLead = "#";

    lOpen = lDefOpen;
    lClose = lDefClose;
    lEscapeOpen = lDefEscapeOpen;
    lEscapeClose = lDefEscapeClose;
    lCommentLead = lDefCommentLead;

    lMacroDefFileName = pPath.dup;

    if ((lMacroDefFileName.length > 0) &&
        (lMacroDefFileName[$-std.path.sep.length..$] != std.path.sep) )
            lMacroDefFileName ~= std.path.sep;

    lMacroDefFileName ~= "build.mdf";


    if (pVerbose && util.fileex.FileExists(lMacroDefFileName) )
        writefln("Build Macro Definition file %s", lMacroDefFileName);
    lMacroDefLines = util.fileex.GetTextLines(lMacroDefFileName, util.fileex.GetOpt.Always);

    foreach(char[] lArg; lMacroDefLines)
    {
        char[][] lOptions;

        lArg = ExpandEnvVar(lArg);
        // Locate any comment text in the line.
        int lPos = std.string.find(lArg, lCommentLead);
        if (lPos != -1)
        {
            // Truncate the line at the comment lead-in character.
            lArg.length = lPos;
        }

        lArg = std.string.strip(lArg);
        if (lArg.length > 1)
        {

            if (lArg.begins("replace ") > 0)
            {
                int lStartPos;
                int lEndPos;
                bool lEndFound;
                int lDepthInc;
                int lDepth;
                int lDataPos;
                char[] lPattern;
                char[] lResolution;

                lArg = std.string.strip(lArg[8..$]);
                if (lOpen in vNestedDelim)
                    lDepthInc = 1;
                else
                    lDepthInc = 0;

                // Format should be <open> <pattern> <close> <open> <resolution> <close>
                lStartPos = 0;
                lEndPos = lStartPos + lOpen.length;
                while( (lEndPos < lArg.length) &&
                       (lArg[lStartPos .. lEndPos] != lOpen)
                      )
                {
                    lStartPos ++;
                    lEndPos ++;
                }
                if (lEndPos >= lArg.length)
                    continue;   // No open delim found.
                lDepth += lDepthInc;
                lDataPos = lEndPos;
                lStartPos = lDataPos;
                lEndPos = lStartPos + lClose.length;
                while( (lEndPos < lArg.length) &&
                       (lArg[lStartPos .. lEndPos] != lClose)
                      )
                {
                    lStartPos ++;
                    lEndPos ++;
                }
                if (lEndPos >= lArg.length)
                    continue;   // No close delim found.
                lPattern = std.string.strip(lArg[lDataPos .. lStartPos]);

                lStartPos = lEndPos;
                lEndPos = lStartPos + lOpen.length;
                while( (lEndPos < lArg.length) &&
                       (lArg[lStartPos .. lEndPos] != lOpen)
                      )
                {
                    lStartPos ++;
                    lEndPos ++;
                }
                if (lEndPos >= lArg.length)
                    continue;   // No open delim found.
                lDepth += lDepthInc;
                lDataPos = lEndPos;
                lStartPos = lDataPos;
                lEndPos = lStartPos + lClose.length;
                while( (lEndPos < lArg.length) &&
                       (lArg[lStartPos .. lEndPos] != lClose)
                      )
                {
                    lStartPos ++;
                    lEndPos ++;
                }
                if (lEndPos > lArg.length)
                    continue;   // No close delim found.
                lResolution = std.string.strip(lArg[lDataPos .. lStartPos]);

                vSubstitutions.length = vSubstitutions.length + 1;
                vSubstitutions[$-1].Pattern = lPattern;
                vSubstitutions[$-1].Resolution = lResolution;
                vSubstitutions[$-1].EscapeOpen = lEscapeOpen;
                vSubstitutions[$-1].EscapeClose = lEscapeClose;

            }
            else if (lArg.begins("delim ") > 0 )
            {

                lArg = std.string.strip(lArg[6..$]);

                lOptions = std.string.split(lArg);
                foreach(char[] lOption; lOptions)
                {
                    if (lOption == "std")
                    {
                        lOpen = lDefOpen;
                        lClose = lDefClose;
                        lEscapeOpen = lDefEscapeOpen;
                        lEscapeClose = lDefEscapeClose;
                        lCommentLead = lDefCommentLead;
                    }
                    else
                    {
                        char[][] lKeyValue;
                        lKeyValue = std.string.split(lOption, "=");
                        switch (lKeyValue[0])
                        {
                            case "open":
                                lOpen = lKeyValue[1];
                                break;
                            case "close":
                                lClose = lKeyValue[1];
                                break;
                            case "escapeopen":
                                lEscapeOpen = lKeyValue[1];
                                break;
                            case "escapeclose":
                                lEscapeClose = lKeyValue[1];
                                break;
                            case "comment":
                                lCommentLead = lKeyValue[1];
                                break;
                            default:
                                if (pVerbose)
                                    writefln("Bad Macro 'delim' option '%s'",
                                            lOption);
                                // Ignore bad options.
                                break;
                        }
                    }
                }
            }
            else if (vVerbose)
            {
                writefln("Bad configuration command '%s'", lArg);
            }
        }
    }
}

void ProcessBuildConfig(char[] pArg, bool pVerbose)
{
    // From build.exe location
    ProcessOneBuildConfig(pArg, pVerbose, std.path.getDirName(vAppPath));

    // From compiler location
    ProcessOneBuildConfig(pArg, pVerbose, vCompilerPath);

}

void ProcessOneBuildConfig(char[] pArg, bool pVerbose, char[] pPath)
{
    char[] lConfigFileName;
    char[][] lConfigLines;
    bool lFoundGroup;


    lConfigFileName = pPath.dup;

    if ((lConfigFileName.length > 0) &&
        (lConfigFileName[$-std.path.sep.length..$] != std.path.sep) )
            lConfigFileName ~= std.path.sep;

    lConfigFileName ~= "build.cfg";

    if (pVerbose  && util.fileex.FileExists(lConfigFileName))
        writefln("Build Configuration file %s [%s]", lConfigFileName, pArg);

    lConfigLines = util.fileex.GetTextLines(lConfigFileName, util.fileex.GetOpt.Always);
    if (pArg.length == 1)
        lFoundGroup = true;
    else
    {
        lFoundGroup = false;
        pArg = pArg[1..$];
    }

    foreach(char[] lArg; lConfigLines)
    {
        lArg = ExpandEnvVar(lArg);
        if (!lFoundGroup)
        {
            if (lArg == "[" ~ pArg ~ "]")
            {
                lFoundGroup = true;
                continue;
            }
        }

        if (lFoundGroup)
        {
            // Locate any comment text in the line.
            int lPos = std.string.find(lArg, "#");
            if (lPos != -1)
            {
                // Truncate the line at the '#' character.
                lArg.length = lPos;
            }

            lArg = std.string.strip(lArg);
            if (lArg.length > 1)
            {
                if ((lArg[0] == '[') && (lArg[$-1] == ']'))
                    break; // Don't process any more lines.

                if (pVerbose)
                    writefln("Build Configuration file arg: %s", lArg);


                while ((lPos = find(lArg, "{Group}")) != -1)
                {
                    lArg = lArg[0..lPos] ~ pArg ~ lArg[lPos + 7 .. $];
                }

                if (lArg.begins("CMDLINE=") > 0)
                {
                    int lStartPos;
                    int lEndPos;
                    bool lEndFound;
                    lArg = std.string.strip(lArg[8..$]);
                    lStartPos = 0;
                    lEndPos = 0;
                    lEndFound = false;
                    while(lEndPos < lArg.length)
                    {
                        if (!lEndFound)
                        {
                            if (lArg[lEndPos] == ' ')
                            {
                                lEndFound = true;
                            }
                        }
                        else
                        {
                            if (find("-+@", lArg[lEndPos..lEndPos+1]) != -1)
                            {
                                GatherOneArg( lArg[lStartPos..lEndPos] );
                                lStartPos = lEndPos;
                                lEndFound = false;
                            }
                        }
                        lEndPos++;
                    }
                    if (lStartPos != lEndPos)
                        GatherOneArg( lArg[lStartPos..lEndPos] );

                }
                else if (lArg.begins("LIBCMD=") > 0)
                {
                    int lPos;

                    while( (lPos = find(lArg, "\"")) != -1)
                    {
                        lArg = lArg[0..lPos] ~ lArg[lPos+1 .. $];
                    }
                    vLibrarianPath = std.path.getDirName(lArg[7..$]) ~ std.path.sep;
                    vLibrarianPath = util.pathex.CanonicalPath(vLibrarianPath);

                    vLibrarian = std.path.getBaseName(lArg[7..$]);
                    if(vVerbose) {
                        writefln(" librarian path %s",vLibrarianPath);
                        writefln(" librarian is %s",vLibrarian);
                    }
                }
                else if (vVerbose)
                {
                    writefln("Bad configuration command '%s'", lArg);
                }
            }
        }
    }
}

/* ============= The User Manual ======================================
__ /topic Introduction
__ /info
__ This is a utility to build an application using the D programming language.
__ It does this by examining the source file supplied on the command line
__ to work out its dependant files, and then determines which need to be
__  compiled and which need to be linked to create the executable.
__
__ Alternatively, it can be used to create a Library file rather than an
__  executable.
__
__ The aim of the utility is to help remove the need for /i make files
__  or similar devices.

__ /topic Pragma
__ /info
__ The /b build utility supports the use of various pragma statements.
__ A pragma is a special statement embedded in the source code that
__  provides information to tools reading the source code.
__
__ They take the forms ...
__ /code
__       /b"pragma(" ~<name~> /b");"
__       /b"pragma(" ~<name~> /b"," ~<option~> [ /b","~<option~>] /b");"
__ /endcode
__
__ If the D compiler doesn't recognise the pragma, it will fail. So to
__  'hide' them from the compiler, you need to wrap them in a /b version
__  block. All the pragmas used by this utility need to be enclosed in
__  a /i build version.
__
__ Example:
__ /code
__       version(build) { pragma(nolink); }
__ /endcode

__ /topic Pragma
__ /pragma link
__ /desc This nominates one or more libraries that are required to be linked in.
__ If your applications needs code from a library to be linked in, rather than
__  supplying source code, you can tell /i build which libraries are needed.
__ This can happen when using a library provided by a third-party.
__
__ Example 1:
__ /code
__       /// This app needs the MyGUI.lib library to be used.
__       version(build) { pragma(link, MyGUI); }
__ /endcode
__
__ Example 2:
__ /code
__       /// This app needs the a DB library and TCP library to be used.
__       version(build) { pragma(link, EuDB, TCP4Win); }
__ /endcode

__ /topic Pragma
__ /pragma nolink
__ /desc This identifies that the current module is not to be be linked in.
__ Normally, each object file created by the compiler is linked in, but
__  if the supplied source file is just a stub for code which is externally
__  defined in a library, then you do not need the 'stub' object file.
__
__ Example:
__ /code
__       version(build) { pragma(nolink); }
__ /endcode

__ /topic Pragma
__ /pragma include
__ /desc This identifies a required file which is not otherwise imported.
__ In some applications, especially ones converted over from C, it is
__ possible that the file on the /i Build command line does not directly
__ or indirectly import a required file. In those situations, you can
__ use this pragma to tell build to include it in the compilation checking
__ process.
__
__ Example:
__ /code
__       // Tell 'build' that prime.d must be included (it contains the main function.)
__       version(build) { pragma(include, "prime"); }
__ /endcode

__ /topic Pragma
__ /pragma target
__ /desc This identifies the basename of the target file.
__ By default, the target name is based on the first file on the command line.
__ But if you include this pragma, the name identified in the pragma becomes
__ the default name. In either case, the -T switch overrides the default name.
__
__  If two or more /i target pragmas are found, the first one is the one
__  that is used and the others are ignored (though mentioned if in
__  verbose mode).
__ Example:
__ /code
__       //// Tell 'build' to create WhizzBang.exe
__       version(build) { pragma(target, "WhizzBang"); }
__ /endcode

__ /topic Pragma
__ /pragma build
__ /desc This identifies a file that needs an external program to build it.
__ Some applications need to link in object files created by C source, or
__ by a /i resource compiler, or whatever. This pragma identifies a file
__ that needs to be linked in, but cannot be created by the D compiler. The
__ format is /n
__ /code
__     pragma(build, "FILENAME" [, "OPTIONS"] ...);
__ /endcode
__ where /i FILENAME is either the file to link in, or the file to use when
__ creating the file to link in. For example, if you had a Windows resource file
__ that needed to be compiled, you could code the pragma as either ...
__
__ /code
__     //// Compile the images into a resource obj and add images.res to linker.
__     pragma(build, "images.rc");
__ /endcode
__ or /n
__ /code
__     //// Compile the images into a resource obj and add images.res to linker.
__     pragma(build, "images.res");
__ /endcode
__
__ The first example specifies the source file to be passed to the resource
__ compiler and the second example specifies the output of the resource compiler.
__ In either case, this utility uses the rules in a /"Rule Definition File" to
__ decide what to do.
__
__ The utility searches for the /i"FILENAME" in the currently defined 'import'
__ paths.
__
__ The /i OPTIONS can be included if you need to pass any special values
__ to the external tool. There can be any number of these, but each one must
__ take the form ~<KEYWORD>=~<VALUE>. For example /n
__ /code
__       pragma(build, "dbapi.c", "COPT=-wc -x", "HDR=abc.hp");
__ /endcode
__ The OPTIONS values are used as replacement text for token in the
__ Rule Definition File's 'tool' specification. In the example above,
__ the tokens {COPT} and {HDR}, if found in the 'tool' line, would be
__ replaced with /i"-wc -x" and /i"abc.hp" respectively.
__
__ The output file to the external tool is checked to see if it is still
__ up to date.
__
__ The output file from the external tool is added to the linkage set.
__
__ All such external programs are run before the D compiler is invoked.
__
__ Example:
__ /code
__       //// Tell 'build' that it needs to call an external program
__       //// to build an up-to-date version of 'images.rc'
__       version(build) { pragma(build, "images.rc"); }
__ /endcode

__ /topic Rule Definition File
__ /info
__ Defines the rules used to build non-D source files.
__ This is not as advanced as the typical /b makefile abilities and is
__ only useful for simple situations. Specifically, it only caters for
__ applications that take a single input file to create the required
__ output file.
__
__ The default Rule Definition File is called "default.rdf". It is found by
__ first looking in the directory where the utility is installed, and if
__ it is not found there, the directories in the PATH environment symbol
__ are scanned.
__
__ The default RDF can be overridden by the -RDF command line switch.
__
__ Each rule is defined in a set of four mandatory lines ... /n
__ /li rule=~<name>
__ /li source=~<extention>
__ /li target=~<extention>
__ /li tool=~<command pattern>
__
__ Any other line in the file, not starting with one of the above keywords
__ is simply ignored (treated as comments).
__
__ The /i"tool" line can contain special tokens. These are keywords enclosed
__ in braces. At compile time, the tokens are replaced by replacement text
__ taken from the pragma(build) statement. There are two predefined tokens
__ that can be used. /b"{@IN}" is replaced by the name of the input file,
__ and /b"{@OUT}" is replaced by the name of the output file.
__
__ /code
__  ----- Windows Resource Compiler -----
__  This uses ` pragma(/b'build', ~<sourcefile>.rc); `
__  -------------------------------------
__  rule=Resources
__  source=rc
__  target=res
__ tool=rc //r {@IN} //fo {@OUT}
__
__  -------------------------------------
__  DMD C linkage
__  This uses ` pragma(build, ~<sourcefile>.c, "COPT=~<options>", "HDR=~<whatever>"); `
__  -------------------------------------
__ rule=DMD C/C++
__  source=c
__  target=obj
__  tool=dmc -c {COPT} {@IN} -o{@OUT} {HDR}
__
__ /endcode

__ /topic Pragma
__ /pragma export_version
__ /desc This allows you to set a global version identifier.
__ DMD allows you to set a version identifier in your code, but the scope
__ of that is only for the module it is set in. This pragma gives you the
__ ability to declare a version identifier which is applied to all modules
__ being compiled, and not just the 'current' module.
__
__ Example:
__ /code
__     version(build) pragma(export_version, Unix);
__     version(build) pragma(export_version, Limited);
__ /endcode
__ These lines will cause the compiler to have these version identifiers
__ added to the command line switches, thus making them effectively global.
__
__ You can list more than one identifier on the pragma statement ...
__ /code
__     version(build) pragma(export_version, Unix, Limited);
__ /endcode

__ /topic Pragma
__ /pragma build_def
__ /desc This supplies an option to be placed in an OptLink definition file.
__ You can have /i build create a customised OptLink definition file
__ by coding as many /i build_def pragmas as required. However, /i build will
__ only allow the first of each type of Definition File command to be used. THis
__ means that if you code ... /n
__ /code
__       pragma (build_def, "EXETYPE DOS");
__       pragma (build_def, "EXETYPE NT");
__ /endcode
__ Then the EXETYPE DOS will be used and the 'NT' line ignored. You can use
__ explicit build_def pragmas to override the default ones generated by
__  /i build for Windows programs or DLL libraries.
__
__ The syntax for these pragma is /n
__ /code
__           pragma(build_def, ~<QUOTED_STRING~> );
__ /endcode
__
__ Example:
__ /code
__   version(build) {
__     pragma (build_def, "VERSION 1.1");
__     version(DOS) {
__       pragma (build_def, "EXETYPE DOS");
__     }
__     version(WIN) {
__       pragma (build_def, "EXETYPE NT");
__       pragma (build_def, "SUBSYSTEM WINDOWS,4.0");
__     }
__   }
__ /endcode
__
__ You can supply /b any text string and it is used verbatim. There is no
__  restrictions on what you can include in the pragma.

__ /topic Auto Build Number
__ /info
__ Automatically Incremented Build Numbering for Modules
__ You can optionally specify that the /i build utility automatically
__ increments a build number for any module. You do this by supplying
__  a single line in your source code in the form ... /n
__ /code
__       private import /b"~<modulename~>"_bn;
__ /endcode
__ Typically this line is placed immediately after the source file's
__ /b module statement, but that is not strictly necessary.
__
__ By having this line in your code, /i build then uses a file with the
__  suffix "_bn.d" to maintain
__ the current build number for the ~<modulename~>. Note that /i build
__  will create this file if it doesn't exist. Also note that
__  /i build will update this file whenever the module file has been
__  updated or its object file needs to be rebuilt, so you really
__ shouldn't modify it manually. Any manual changes will be deleted
__  by /i build. /n
__
__ Example: /n
__ If your module is called "parser.d" you would have the lines ...
__ /code
__   module parser;
__   private import parser_bn;
__ /endcode
__
__ You can access the build number from within the module thus ... /n
__
__ /code
__           writefln("Build number is %d", auto_build_number);
__ /endcode
__ You can access the build numbers of other modules in you application
__  by importing the appropriate file and prefixing the references with
__  the module names. /n
__ Example:
__ /code
__       module parser;  /// This module's name.
__       private import parser_bn;   /// This module's B/N
__       private import tokenizer_bn; /// Another module's B/N
__       . . .
__       writefln("Builds...");
__       writefln("  Parser %d", parser_bn.auto_build_number);
__       writefln("  Tokens %d", tokenizer_bn.auto_build_number);
__ /endcode
__
__ The "_bn.d" file created by /i build for this module would look like ... /n
__ /code
__   module parser_bn;
__   /// This file is automatically maintained by the BUILD utility,
__   /// Please refrain from manually editing it.
__   long auto_build_number = 77;
__ /endcode
__ Of course the number /i 77 is just an example. This number would actually
__  start at 1 and increment whenever /i build needed to create a new object
__ file for the module.

__ /topic Command Line
__ /info
__ The /i build utility supplies the source file and optional switches via the command line.
__ The typical format of the /i build utility's command line is ... /n
__ /code
__       /b build ~<sourcefile~> [ ~<switches~> ] [ ~<otherfiles~> ]
__ /endcode
__ You can however specify the files and switches in any order you like.
__
__ Normally you would only supply a single file name to /i build, that being
__  the source file of the top-level file in the application. Typically
__  the one that has the 'main()' or 'WinMain()' function, though this is
__  not strictly necessary.
__
__ If you run the utility without any files and without any switches, it
__  display a /i help screen with some details about the switches. That looks
__  like this ...
__ /code
__   Path and Version : C:\Program Files\build.exe v3.6(339)
__   Usage: build sourcefile [options objectfiles libraries]
__     /b sourcefile D source file
__     /b"-v"         Verbose (passed through to D)
__     /b"-V"         Verbose (NOT passed through)
__     /b"-names"     Displays the names of the files used to build the target.
__     /b"-DCPATH~<path~> ~<path~>" is where the compiler has been installed.
__                Only needed if the compiler is not in the system's
__                PATH list. Used if you are testing an alternate
__                version of the compiler.
__     /b"-CFPATH~<path~> ~<path~>" is where the D config file has been installed.
__     /b"-full"      Causes all source files, except ignored modules,
__                 to be compiled.
__     /b"-link"      Forces the linker to be called instead of the librarian.
__                 (Only needed if the source files do not contain
__                  main/WinMain)
__     /b"-nolink"    Ensures that the linker is not called.
__                 (Only needed if main/WinMain is found in the source
__                  files and you do NOT want an executable created.)
__     /b"-lib"       Forces the object files to be placed in a library.
__                 (Only needed if main/WinMain is found in the source
__                  files AND you want it in a library instead of
__                  an executable.)
__     /b"-nolib"     Ensures that the object files are not used to form
__                 a library.
__                 (Only needed if main/WinMain is not found in the source
__                  files and you do NOT want a library.
__     /b"-obj"       This is the same as having both -nolink and -nolib
__                 switches on the command line. It is just a shorthand.
__     /b"-allobj"    Ensures that all object files are added to a
__                 library.
__                 (Normally only those in the same directory are added.)
__     /b"-cleanup"   Ensures that all object files created during the run
__                 are removed at the end of the run, plus other work files.
__     /b"-clean"     Same as -cleanup
__     /b"-test"      Does everything as normal except it displays the commands
__                   instead of running them. Also, the auto-build-numbers are
__                   not incremented.
__     /b"-RDF~<path~>"  Overrides the default Rule Definition File
__     /b"-gui[:x.y]" Forces a GUI application to be created. The optional
__                 :x.y can be used to build an application for a
__                 specific version of Windows. eg. -gui:4.0
__                 (Only needed if WinMain is not found in the source files
__                 or if you wish to override the default Windows version)
__     /b"-dll"       Forces a DLL library to be created.
__                       (Only needed if DllMain is not found in the source files.
__     /b"-LIBOPT~<opt~>" Allows you to pass ~<opt~> to the librarian
__     /b"-LIBPATH=~<pathlist~>" Used to add a semi-colon delimited list
__                 of search paths for library files.
__     /b"-X~<module~>" Packages and Modules to ignore (eg. -Xmylib)
__     /b"-M~<module~>" Packages and Modules to notice (eg. -Mphobos)
__     /b"-T~<targetname~>" The name of the target file to create. Normally
__                 the target name istaken from the first or only name
__                 of the command line.
__     /b"-R~<y|n~>" Indicates whether to use a response file or command line
__                    arguments with the compiler tools.
__                  /b"-Ry" will cause a response to be used.
__                  /b"-Rn" will cause command line arguments to be used.
__                  /b"-R" will reverse the current usage.
__                   Note that the default for Windows DigitalMars is to use a response file
__                   but for all other environments it is to use command line arguments.
__     /b"-info"      Displays the version and path of the Build application
__     /b"-help"      Displays the full 'usage' help text.
__     /b"-h"         Same as /-help, displays the full 'usage' help text.
__     /b"-?   "      Same as /-help, displays the full 'usage' help text.
__     /b"-silent"    Avoids unnecessary messages being displayed.
__     /b"-noautoimport" Turns off the automatic addition of source paths
__                    to the list of Import Roots.
__     /b"-run~<param~>"   If the link is successful, this will cause the
__                    executable just created to run. You can give it
__                    run time parameters. Anything after the '-run' will
__                    placed in the program's command line. You will need
__                    to quote any embedded spaces.
__     /b"-od~<path~>"  Nominate the directory where temporary (work) files
__                    are to be created. By default they are created in
__                    the same directory as the target file.
__     /b"-nodef"     Prevents a Module Definition File being automatically created.
__                 This will override any pragma(/build_def) statements in the
__                 source code.
__     /b"[...]"      All other options, objectfiles and libraries are
__                 passed to the compiler
__ /endcode
__
__ See also /"Response File"

__ /topic Switches
__ /info
__ /i Build supports a number of command line switches to control its default assumptions.


__ /topic switches
__ /switch -nodef
__ /desc Prevent a Module Definition File being created.
__ Normally, the utility will automatically create a Module Definition File
__ for the linker. This will typically contain the EXETYPE command and any other
__ commands as specified in any pragma(/build_def) statements found in the source files.
__ If you specify this switch, you will have to provide your own Module Definition
__ File or its equivalent on the command line.

__ /topic switches
__ /switch -v
__ /desc Set /i verbose mode on for both /i build and for the compiler

__ /topic switches
__ /switch -V
__ /desc Set /i verbose mode on for just /i build and not for the compiler

__ /topic switches
__ /switch -names
__ /desc Displays the names of the files used in building the target.

__ /topic switches
__ /switch -DCPATH
__ /desc  Identifies where the compiler has been installed.
__ Normally, /i build scans the PATH environment symbol to find where the
__  D compiler is located. However, if you need to use the compiler from
__  a different location, you would use this switch to tell /i build where
__  it is.
__
__ /b"Note:" If you use this switch, and the current CFPATH value is
__ identical to the current DCPATH value, then this switch will change
__ both to the new value. The assumption is that the configuration
__ file is in the same directory as the compiler. If this is not the case,
__ you will also need to use the /-CFPATH switch.
__
__ Example:
__ /code
__   -DCPATHc:\old\dmd\bin
__ /endcode

__ /topic switches
__ /switch -CFPATH
__ /desc Identifies where the D config file has been installed.
__ Normally, /i build looks in the same place that the compiler is
__  installed in, but if you need to use a different configuration
__  path from that, you would use this switch to tell /i build where
__  it is.
__
__ Example:
__ /code
__   -CFPATHc:\myproject\configs
__ /endcode

__ /topic switches
__ /switch -full
__ /desc Causes all source files, except ignored modules, to be compiled.
__ Normally, /i build only compiles a source file if it's object file
__ is out of date or missing. This switch forces all source files to be
__  recompiled, even if not strictly required.
__
__ /b "Note:" Modules in the /i ignore list are still ignored though.

__ /topic switches
__ /switch -link
__ /desc Forces the linker to be called instead of the librarian.
__ Normally, if /i build does not find a main() or WinMain() function
__  in the source files, it creates a library to contain the object
__ files. But when this switch is used, /i build will attempt to
__  create an application by calling the linker rather than the librarian.
__
__ You would typically use this if the main() function was being supplied
__ from an existing library file rather than one you your source files.

__ /topic switches
__ /switch -run
__ /desc Runs a program after successful linking.
__ If the link is successful, this will cause the
__ executable just created to run. You can give it
__ run time parameters. Anything after the '-run' will
__ placed in the program's command line. You will need
__ to quote any embedded spaces.
__
__Example:
__/code
__     -run"abc.de second"
__/endcode

__ /topic switches
__ /switch -od
__ /desc Nominates the temporary file directory.
__ By default, the utility creates any work files in the same directory
__ as the target file. You can use this switch to nominate an alternative
__ location for these files. The directory is created if it doesn't exist.
__
__ For the DigitalMars compiler, this also specifies the top-level
__ location where object files are created. By default, object files
__ are created in the same directory as the corresponding source file.
__
__ Example:
__ /code
__         -odC:\temp\workarea
__ /endcode

__ /topic switches
__ /switch -nolink
__ /desc  Ensures that the linker is not called.
__ Normally, if /i build finds a main() or WinMain() function, it
__ tries to create an application by calling the linker. If you use
__  this switch however, the linker will not be called.
__
__ You would typically do this if you wanted to create a library that
__ stores your 'main' function in it, in which case you'd also use the /-lib
__  switch. But you could just create a set of object files without linking
__  them with this switch.

__ /topic switches
__ /switch -lib
__ /desc Forces the object files to be placed in a library.
__ Normally, if /i build finds a main() or WinMain() function, it
__ tries to create an application by calling the linker. But if you
__  use this switch, the librarian is called instead of the linker.
__
__ You would typically do this if you wanted to create a library that
__ stores your 'main' function in it, in which case you'd also use the /-nolink
__ switch.

__ /topic switches
__ /switch -obj
__ /desc Shorthand for using both /"-nolink" and /"-nolib" switches.
__ This is literally the same has if you had placed both -nolink and
__ -nolib on the command line. As this is a common way to just compile
__ modules to get the object files for them, without doing anything else,
__ a neat shorthand is available.

__ /topic switches
__ /switch -nolib
__ /desc Ensures that the object files are not used to form a library.
__ Normally, if /i build does not find a main() or WinMain() function,
__  it calls the librarian to create a library for all your object files.
__ But if you use this switch, the librarian is not called.
__
__ Typically, you would do this if you just wanted to create one or more
__  object files and do nothing else.

__ /topic switches
__ /switch -allobj
__ /desc Ensures that all object files are added to a library.
__ Normally, /i build will only create a library using the object files
__  that are in the same directory as the new library. You would use
__  this switch if you wanted all object files created by this build
__ session to be included in the library.

__ /topic switches
__ /switch -cleanup
__ /desc Ensures that all working files created during the run are removed.
__ Normally, /i build does not delete any object files or working files
__ when it finishes a session. You can use this switch to have /i build
__  clean up after itself. This will remove all object files created
__  in this run, plus any .rsp, .lsp, .bak, .map, .lst, .def files.
__
__  /b Note: This can also be supplied as /b"-clean" as an alias.

__ /topic switches
__ /switch -LIBOPT
__ /desc Allows commandline options to be passed to the librarian.
__ Example:
__ /code
__     (Set the page size to 32Kb)
__     -LIBOPT-p32
__
__     (Embedded spaces enclosed in quotes.)
__     "-LIBOPT -l -i"
__ /endcode

__ /topic switches
__ /switch -LIBPATH
__ /desc Used to add a list of search paths for library files.
__This might be used when you don't want to permanently update the
__ standard search paths.
__ Example:
__ /code
__     -LIBPATH=c:\mylibs;d:\3rdparty;c:\lib\debuglibs
__ /endcode

__ /topic switches
__ /switch -R
__ /desc Determines if the compiler tools use a response file or not.
__ For /i DigitalMars tools in the Windows environment, a response file is
__  the default, but for /i other tools and other operating systems the
__  default is to use command line arguments.
__
__ The switch has three formats: /n
__  -Ry to use a response file /n
__  -Rn to use command line arguments /n
__  -R to reverse the current setting.
__
__ The use of a response file is only really needed when the command line
__ arguments are going to be more than the operating system can handle
__ on a single command line. However, it is always a safe option, so if
__ in doubt you may as well use it.
__
__ A response file contains all the arguments that would have gone on the
__ command line. They are arranged as one argument per line.
__
__ Not all tools respect the response file idea however the Windows based
__  DigitalMars tools do understand it.
__
__  Without a response file the compiler might be invoked thus:
__ /code
__   dmd -op -release appmain.d somemod.obj -IC:\DLibs
__ /endcode
__ But with a response file, these arguments are first written out to
__ a text file (the response file) and the compiler is invoked :
__ /code
__   dmd @appmain.rsp
__ /endcode
__ The /i "appmain.rsp" would contain the lines :
__ /code
__   -op
__   -release
__   appmain.d
__   somemod.obj
__   -IC:\DLibs
__ /endcode

__ /topic switches
__ /switch -test
__ /desc Does a test run only. No compiling, linking or library work is done.
__ This will display the command lines instead of running them. It can be used
__ to see what would happen without actually building anything.

__ /topic switches
__ /switch -RDF
__ /desc Defines a file to override the default /"Rule Definition File"
__ The default file is called "default.rdf". If you need to provide
__ an alternate file, you can use this switch.
__
__ Example:
__ /code
__     -RDFmyrules.xyz
__ /endcode

__ /topic switches
__ /switch -dll
__ /desc Forces a DLL library to be created.
__ /b"Note:" This only applies to Windows environment.
__
__ Normally, if /i build finds a DllMain() function it automatically
__  creates a DLL library. However, if you need to force a DLL
__  library, you would use this switch.

__ /topic switches
__ /switch -gui
__ /desc Forces a GUI application to be created.
__ /b"Note:" This only applies to Windows environment.
__
__ Normally, if /i build finds a WinMain() function it automatically
__  creates a GUI application. However, if you need to force a GUI
__  application, you would use this switch.
__
__ This switch can also be used to specify which version of Windows to
__  build the application for. To do this, it takes the format of
__ -gui:X.Y where /i"X.Y" is the Windows version number. Use 4.0
__  for Windows NT, 2000, and ME, and 5.0 for Windows XP.
__
__ By default, /i build uses the version of Windows it is running under.

__ /topic switches
__ /switch -info
__ /desc Displays the version and path of the /i build application

__ /topic switches
__ /switch -help
__ /desc Displays the full text of the Usage information.
__ This has the aliases of "-h" and "-?"

__ /topic switches
__ /switch -silent
__ /desc Prevents unnecessary messages being displayed.
__ Some messages are just informational and under some circumstances they
__ can interfer with reading the output.

__ /topic switches
__ /switch -noautoimport
__ /desc Prevents source file paths from being added to the list of Import Roots
__ By default, for each source file that imports a module, it's path is added
__ to the list of paths that will be searched for module source files. If you
__ do not wish that behaviour, you will need to use this switch. In that case,
__ the compiler will only search the paths specified in the compiler's
__ configuration file, the current directory, and any explicitly added paths
__ on the command line.
__
__ Example:
__ /code
__      build myApp -noautoimport
__ /endcode

__ /topic switches
__ /switch -X
__ /desc Identifies a module or package to ignore
__ Normally, /i build assumes that all imported modules are available
__ to be recompiled if required. You would use this switch if you explictly
__  did not want /i build to recompile a module.
__
__ /b"Note:" The /i Phobos package of modules is automatically ignored. This
__  means that /i build does not try to recompile phobos.
__
__ Example - ignore the 'parser' module (or package)
__ /code
__       -Xparser
__ /endcode

__ /topic switches
__ /switch -M
__ /desc Identifies a module or a package to notice (not ignore)
__ You would use this to name any module that is not part of the target's
__ dependancies, or is one of the /i ignored modules.
__
__ You can use this switch to recompile /i phobos.
__
__ Example (notice the Phobos package):
__ /code
__   -Mphobos
__ /endcode

__ /topic switches
__ /switch -T
__ /desc Identifies the target name to build.
__ Normally, /i build derives the target name from the first file name
__ on the command line, or from the pragma(target) if present.
__ If however, you wish to override that, use this switch.
__
__ /b"Note:" This switch allows the use of a special token, /b"{Target}",
__ which is replaced by the default target name. You can use
__ to create a generic /"Response File" that can be used for building
__ any application.
__
__ Example
__ /code
__       -Ttestapp
__
__       -Ttest_(Target}  # Generate a name based on the default name.
__ /endcode

__ /topic Response File
__ /info
__ A response file is a file that contains command line values.
__
__ You can specify all or any command line values in a
__ response file. Each value appears in its own line in the
__ response file and you reference this file by prefixing
__ its name with an '@' symbol on the command line.
__
__ You may place comments in a response file. A comment starts
__  with a /b'#' character and extends to the end of the line.
__
__ Example:
__ /code
__    build @final
__ /endcode
__ where a file called 'final.brf' contains the command
__ line values, including other response file references.
__
__ Example: (contents of final.brf)
__ /code
__     # This creates a production (release) edition of the app.
__     -T{Target}_release  # Set the name of the executable.
__     -release   # Don't generate runtime checks.
__     -full      # Force compilation of all files.
__     -cleanup   # remove work files when completed
__     -inline    # Allow inlining to occur
__ /endcode
__
__ The command line for your build could then look like ...
__ /code
__     build myapp @final
__ /endcode
__
__ If the response file reference is just a single '@' then
__  build looks for a file called 'build.brf'
__
__ You can any number of response file references on a command line. They
__  are processed in the order they appear.


__ /topic Configuration File
__ /info
__ The utility configuration file is used to specify your default options for the utility.
__
__ Each option is specified in a single text line, but you can have any
__ number of option lines in the configuration file.
__
__ The options take the form of
__ /code
__      ~<COMMAND~>=~<VALUE~>
__ /endcode
__
__ /b"Supported Commands" /n
__ The commands currently supported are ... /n
__ /b CMDLINE /n
__ You can specify all command line switch values with this configuration
__ option. You can also specify multiple switches on the same option line./n
__ /code
__     CMDLINE=-inline -w
__ /endcode
__
__ /b LIBCMD /n
__ This specifies the location and program name for the librarian application
__ you want to invoke when creating libraries. /n
__ /code
__    LIBCMD=D:\Applications\dm\bin\lib.exe
__ /endcode
__
__ /b"Environment Symbol Substitution" /n
__ Before each configuration file option line is processed, it is first checked
__ for any references to Environment symbols. Each reference is replaced by
__ the value of that symbol. References take the form /b"%~<SYMNAME~>%" /n
__ There are two special symbols: /b"@D" is replaced by the compiler's
__ path, and /b"@P" is replaced by the compiler's configuration file's path.
__
__Example: Assuming you had set BUILDOPTS=-w -g
__ /code
__   CMDLINE=%BUILDOPTS%    # Allow switches to be passed via enviroment symbol.
__ /endcode
__ would mean that "-w -g" would be automatically placed on the utility's
__ commandline whenever you ran it.
__
__ /b"Groups" /n
__ It is possible to specify groups of configuration options that are only
__ applied if explictly nominated on the command line. A group starts with
__ a line in the format /b"[~<groupname~>]" where /i groupname can be any
__ text that doesn't include spaces. A group ends on the last line before the
__ next group in the file.
__
__ Example of a Group
__ /code
__    [dbg] # To produce a debug edition of an application.
__    CMDLINE=-unittest
__    CMDLINE=--release
__    CMDLINE=--inline
__    CMDLINE=-g
__    CMDLINE=-w
__    CMDLINE=-full
__    CMDLINE=-T{Target}_{Group}
__ /endcode
__
__ All the options lines before the first group are known as the default
__ options because these are always applied. To apply the options in a group
__ you need to specify which group(s) you want on the command line, or via
__ a /"response file". The group name is prepended with a plus sign on
__ the command line. To apply the 'debug' group in the above example, you
__ place on the command line /b"+dbg"
__
__ /b"Comments" /n
__ You may place comments in a configuration file. A comment starts
__  with a /b'#' character and extends to the end of the line.
__
__ Example:
__ /code
__    build myapp.d +final
__ /endcode
__ where if a group called '[final]' is contained in any configuration
__ file, the options in that group are used.
__
__ Example: (contents of [final])
__ /code
__     # This creates a production (release) edition of the app.
__     [final]
__     CMDLINE=-T{Target}_{Group}  # Set the name of the executable.
__     CMDLINE=-release   # Don't generate runtime checks.
__     CMDLINE=-full      # Force compilation of all files.
__     CMDLINE=-cleanup   # remove work files when completed
__     CMDLINE=-inline    # Allow inlining to occur
__     CMDLINE=--debug*   # Turn off any debug switches
__     CMDLINE=--unittest # Turn off any unittest switches
__     CMDLINE=--w        # Turn off warnings
__     CMDLINE=--g        # Turn off embedded debug symbolic info.
__ /endcode
__
__ You can any number of configuration file group references on a command line. They
__  are processed in the order they appear.

__ /topic To Do
__ /info
__ Ideas that have yet to be implemented.
__ Note that some of these may never be implemented but they are still
__ listed here as a reminder.
__
__ /li Be able to update a library rather than just create libraries.
__ /li Support the concept of a 'Plugin' block of code. /n
__ /code
__  version (build) pragma(plugin, ~<tool>[,delim=xxx],[, ~<parms>] ) { [body] }
__ /endcode
__ This block would be replaced by the stdout data of calling 'tool ~<params>' and
__  sending it the content of ~<body> via stdin.
__ /li Support limited preprocessor capability. /n
__ Something like pragma(macro, ~<preprocessor commands>); and replacing
__ source text of the form @{tokenname}@ with the value of the token.

__ /topic DLL Libraries
__ /info
__ How to create a DLL library using Build.
__ When Build finds a DllMain() function, or the -dll switch is used, it
__ creates a special type of Module Definition file, and sets the target to
__ to be a DLL file. The Module Definition file generated takes the form ...
__ /code
__     LIBRARY "~<targetname>.dll";
__     EXETYPE NT
__     SUBSYSTEM WINDOWS,~<version>
__     CODE PRELOAD DISCARDABLE SHARED EXECUTE
__     DATA PRELOAD SINGLE WRITE
__ /endcode
__
__ You can replace any of these by explicitly coding a pragma(/build_def)
__ with different options than these defaults shown here.
__
__ Also, if you have the program /b implib in your path, it will be run
__ after the DLL file is created, in order to also create a /b".lib"
__ library to interface with the DLL. Build currently supports both the
__ DigitalMars version and the Borland (Inprise) version of /b implib.
*/
