/**************************************************************************

        @file source.d

        Copyright (c) 2005 Derek Parnell

        This software is provided 'as-is', without any express or implied
        warranty. In no event will the authors be held liable for damages
        of any kind arising from the use of this software.

        Permission is hereby granted to anyone to use this software for any
        purpose, including commercial applications, and to alter it and/or
        redistribute it freely, subject to the following restrictions:

        1. The origin of this software must not be misrepresented; you must
           not claim that you wrote the original software. If you use this
           software in a product, an acknowledgment within documentation of
           said product would be appreciated but is not required.

        2. Altered source versions must be plainly marked as such, and must
           not be misrepresented as being the original software.

        3. This notice may not be removed or altered from any distribution
           of the source.

        4. Derivative works are permitted, but they must carry this notice
           in full and credit the original source.


                        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


        @version        Initial version, January 2005
        @author         Derek Parnell


**************************************************************************/

module source;
private import source_bn;    // Build number for this module
version(build) pragma(include, "build");
version(unix)  version = Unix;
version(Unix)  version = Posix;
version(linux)  version = Posix;
version(darwin) version = Posix;

private {
    import util.str;
    import util.fdt;
    import util.pathex;
    import util.fileex;
    import util.macro;

    import std.stdio;
    import std.path;
    import std.ctype;
    import std.file;
    import std.string;

    const {
        version(Windows) {
            char[] ExeExt=`exe`;
            char[] LibExt=`lib`;
            char[] ObjExt=`obj`;
        }

        version(Posix) {
            char[] ExeExt=``;
            char[] LibExt=`a`;
            char[] ObjExt=`o`;
        }

        // Using these as the literals confuse my text editor's
        // 'bracket matching' algorithm.
        static char[] kOpenBrace = "\x7B";
        static char[] kCloseBrace = "\x7D";
        static char[] kOpenParen = "\x28";
        static char[] kCloseParen = "\x29";

    }

    bool [char[]] AttributeBlocks;
    bool mainFound;
    bool mainGUI;
    bool mainDLL;
    bool[char[]] vActiveVersions;
    long         vVersionLevel = 0;
    bool[char[]] vActiveDebugs;
    long         vDebugLevel = 0;

}

public {

    char[] function(char[] pPath) AddRoot;
    void function(char[] pArg) AddCompilerArg;
    char[][] function () GetImportRoots;
    bool function() AutoImports;
    void function(char[] pPath) AddTarget;
    void function(char[] pPath) AddLink;
    void function(char[] pPath, char[][] pOpts) AddExternal;
    void function(char[] pText, bool pReplace=false) AddBuildDef;
    char[][] function( ) ModulesToIgnore;
    char[] function(char[] pFile) GetFullPathname;
    char[] function() GetObjWritePath;
    Substitute[] function() GetMacros;

    bool mVerboseMode = false;
    bool mMacroInput = false;

}
    debug(1)
    {
        private import std.stdio;
    }

static this()
{
    AttributeBlocks["private"] = true;
    AttributeBlocks["package"] = true;
    AttributeBlocks["protected"] = true;
    AttributeBlocks["public"] = true;
    AttributeBlocks["export"] = true;
    debug(1)
    {
    writefln("util.str build #%d", auto_build_number);
    }
}

public class Source {
    private static Source[char[]] SourceIndex;

    private {
    char[] mFileName;
    char[] mModuleName;
    char[] mObjectName;
    FileDateTime mObjectTime;     /* time of object file, must be newest, otherwise compile */
    FileDateTime mDependantsTime;   /* time of newest of all dependencies */
    FileDateTime mFileTime;
    bool mHasBeenSearched;
    bool mNoLink;
    bool mIgnore;
    int mBuildNumber;
    char[][] mReferencedImports;
    bool[char[]] mActiveVersions;
    long         mVersionLevel = 0;
    bool[char[]] mActiveDebugs;
    long         mDebugLevel = 0;
    }

    public {
    // FileName
    char[] FileName () { return mFileName.dup; }

    // DependantsTime
    FileDateTime DependantsTime() { return mDependantsTime; }

    // ObjectsTime
    FileDateTime ObjectsTime() { return mObjectTime; }

    // FilesTime
    FileDateTime FilesTime() { return mFileTime; }

    // ModuleName
    char[] ModuleName() { return mModuleName.dup; }
    void ModuleName(char[] pValue) { mModuleName = pValue.dup; }

    // ObjectName
    char[] ObjectName() { return mObjectName.dup; }
    void ObjectName(char[] pValue) { mObjectName = pValue.dup; }

    // NoLink
    bool NoLink() { return mNoLink; }

    // Ignore
    bool Ignore() { return mIgnore; }

    // BuildNumber
    int BuildNumber() { return mBuildNumber; }

// --------------------------------------------------------------------
    this(char[] pFileName)
// --------------------------------------------------------------------
    {
        char[] lObjectName;

        mReferencedImports.length = 0;
        mHasBeenSearched = false;
        mNoLink = false;
        mIgnore = false;
        mFileName = pFileName;
        mModuleName = FileToModulename(pFileName);
        lObjectName = addExt(pFileName,ObjExt);
        mObjectName = GetFullPathname(lObjectName);

        version(DigitalMars)
        {
        /* If a specific path for object files has been supplied
           on the command line, and the current source file is
           relative to the current directory, then we need to
           prepend the specified object location path.
        */
        if (GetObjWritePath().length != 0)
        {
            if (mObjectName == lObjectName)
            {
                mObjectName = GetObjWritePath() ~ mObjectName;
                util.pathex.MakePath(mObjectName);
            }
        }
        }

        mBuildNumber = -1;

        mFileTime = util.fdt.GetFileTime(pFileName);
        mDependantsTime  = mFileTime;

        mObjectTime   = util.fdt.GetFileTime(mObjectName);
        UpdateDependantTime(mObjectTime);

        if(mVerboseMode) {
            writefln("Time %s for %s",mFileTime.toString(),mFileName);
            writefln("Time %s for %s",mObjectTime.toString(),mObjectName);
        }

        // Store this instantiation in the list of known files.
        SourceIndex[pFileName] = this;

        search();
    }

    void search() {
        char[] [] lFiles;
        FileDateTime lCurFileTime;
        bool lCanUse;
        int lTextPos = 0;
        char[] lFileText;

        if(mHasBeenSearched) {
            return;
        }
        else {
        if(mVerboseMode) {
          writefln("Scanning %s", util.pathex.AbbreviateFileName(mFileName));
        }

        // Grab the original text.
        lFileText = GetText(mFileName);

        // Check if any macro processing is required.
        if (mMacroInput && (std.path.getExt(mFileName) != "d"))
        {
            if (util.macro.RunMacro(lFileText, GetMacros()))
            {
                char[] lNewFile;
                lNewFile = std.path.addExt(mFileName, "d");
                std.file.write(lNewFile, lFileText);
                mHasBeenSearched = true;
                mIgnore = true;
                if(mVerboseMode) {
                  writefln("Macro output %s", lNewFile);
                }
                new Source(lNewFile);
                return;
            }
        }

        // Extract all the module references.
        ProcessTokens(lFileText, lTextPos);

        lCurFileTime = new FileDateTime();
        // Examine each extracted module file.
        if ( (mReferencedImports.length > 0) && (AutoImports() == true))
        {

            AddRoot( std.path.getDirName(util.pathex.CanonicalPath(mFileName, false)) );
        }
        foreach(char[] lNextFile; mReferencedImports) {
            if(lNextFile in SourceIndex) {
                // Known file so just grab its mod time
                lCurFileTime = SourceIndex[lNextFile].DependantsTime;
            } else {
                // Ignore specified modules as we assume they are in the lib.
                lCanUse = true;
                if (!(ModulesToIgnore is null)) {
                    foreach(char[] lNextModule; ModulesToIgnore() ){
                        int lFindPos;
                        char[] lType = "package";
                        version(Windows) {
                            char[] lLowerFile;
                            char[] lLowerMod;
                            lLowerFile = std.string.tolower(lNextFile);
                            lLowerMod = std.string.tolower(lNextModule);
                            version(Windows)
                            {
                                // Just in case a Windows user used unix style separators.
                                lLowerMod = std.string.replace(lLowerMod, `/`, `\`);
                            }

                            // Check for package name
                            lFindPos = find(lLowerFile,
                                                    std.path.sep ~ lLowerMod ~ std.path.sep);
                            if (lFindPos == -1) {
                                // It might be a package at the current module level.
                                lFindPos = find(lLowerFile,
                                                    lLowerMod ~ std.path.sep);
                                if (lFindPos != 0)
                                    lFindPos = -1;
                            }
                            // Check for module in subdirectory
                            if (lFindPos == -1) {
                                lType = "module";
                                lFindPos = find(lLowerFile,
                                                    std.path.sep ~ lLowerMod ~ ".d");
                            }
                            // Check for module in current directory
                            if (lFindPos == -1)
                            {
                                lType = "module";
                                if (lLowerFile == lLowerMod ~ ".d")
                                    lFindPos = 0;
                            }
                        }
                        version(Posix) {
                            // Check for package name
                            lFindPos = find(lNextFile,
                                                    std.path.sep ~ lNextModule ~ std.path.sep);
                            if (lFindPos == -1) {
                                // It might be a package at the current module level.
                                lFindPos = find(lNextFile,
                                                    lNextModule ~ std.path.sep);
                                if (lFindPos != 0)
                                    lFindPos = -1;
                            }
                            // Check for module in subdirectory
                            if (lFindPos == -1) {
                                lType = "module";
                                lFindPos = find(lNextFile,
                                                    std.path.sep ~ lNextModule ~ ".d");
                            }
                            // Check for module in current directory
                            if (lFindPos == -1) {
                                lType = "module";
                                if (lNextFile == lNextModule ~ ".d")
                                    lFindPos = 0;
                            }
                        }

                        if( lFindPos >= 0) {
                            if (mVerboseMode)
                                writefln("Ignoring %s (%s: %s)", lNextFile, lType, lNextModule);

                            lCanUse = false;
                            break;
                        }
                    }
                }
                if (lCanUse){
                    // Not known yet, so add it and grab its mod time
                    lNextFile = GetFullPathname(lNextFile);
                    lCurFileTime = (new Source(lNextFile)).DependantsTime;
                }

            }

            // Ensure we get the most recent mod time.
            UpdateDependantTime (lCurFileTime);
        }

        mHasBeenSearched = true;
        }
    }

    void UpdateDependantTime(FileDateTime pDateTime)
    {
        if (pDateTime > mDependantsTime)
        {
            if (mVerboseMode)
            {
                writefln("Updating %s dependants time from %s to %s",
                    mFileName,
                    mDependantsTime.toString(),
                    pDateTime.toString()
                    );
            }
            mDependantsTime = pDateTime;
        }
    }

    static bool WasMainFound() { return mainFound; }
    static void WasMainFound(bool pValue) { mainFound = pValue; }

    static bool WasMainGUI() { return mainGUI; }
    static void WasMainGUI(bool pValue) { mainGUI = pValue; }

    static bool WasMainDLL() { return mainDLL; }
    static void WasMainDLL(bool pValue) { mainDLL = pValue; }

    bool IncrementBuildNumber()
    {
        char[] lFileName;
        Source lBNSource;
        char[] lBaseDir;

        if (mBuildNumber < 0)
            return false;
        else {
        lFileName = ModuleToFilename(mModuleName ~ "_bn");
        lBaseDir = std.path.getDirName(mFileName);
        if (lBaseDir.length > 0 && std.path.getDirName(lFileName).length == 0)
            lFileName = std.path.getDirName(mFileName) ~
                        std.path.sep ~ lFileName;

        lFileName = util.pathex.AbbreviateFileName (
                        util.pathex.CanonicalPath(lFileName, false));
        if (std.file.exists(lFileName))
        {
            std.file.remove(lFileName);
        }

        mBuildNumber++;

        std.file.write(lFileName,
            std.string.format(
            "module %s;\n"
            "// This file is automatically maintained by the BUILD utility,\n"
            "// Please refrain from manually editing it.\n"
            "long auto_build_number = %d;\n",
                    FileToModulename(lFileName),
                    mBuildNumber)
            );

        if (! (lFileName in SourceIndex) )
        {
            lBNSource = new Source(lFileName);
            lBNSource.search();
        }
        else
        {
            lBNSource = SourceIndex[lFileName];
        }
        // Force it to be compiled by pretending it hasn't got an OBJ file.
        lBNSource.mObjectTime = new FileDateTime();
        return true;
        }
    }

    } // End public

    //----------------------------------------------------------
  private {

    // -------------------------------------------
    void ActivateVersion(char[] pID)
    // -------------------------------------------
    {
        if (pID.length > 0)
        {
            if (std.ctype.isdigit (pID[0]))
                mVersionLevel = atoi(pID);
            else
                mActiveVersions [ pID ] = true;
        }
    }

    // -------------------------------------------
    bool IsActiveVersion(char[] pID)
    // -------------------------------------------
    {
        if (pID.length == 0)
            return false;

        if (std.ctype.isdigit(pID[0]))
        {
            long lLevel = atoi(pID);
            if (vVersionLevel > mVersionLevel)
                return lLevel <= vVersionLevel ? true : false ;
            else
                return lLevel <= mVersionLevel ? true : false ;
        }
        else
        return (pID in vActiveVersions) != null ? true :
                 (pID in mActiveVersions) != null ? true : false ;
    }

    // -------------------------------------------
    void ActivateDebug(char[] pID)
    // -------------------------------------------
    {
        static const char[] lDefaultLevel = "1";
        if (pID.length == 0)
            pID = lDefaultLevel;

        if (std.ctype.isdigit(pID[0]))
            mDebugLevel = atoi(pID);
        else
            mActiveDebugs [ pID ] = true;

    }

    // -------------------------------------------
    bool IsActiveDebug(char[] pID)
    // -------------------------------------------
    {

        if (pID.length == 0)
            pID = "1";

        if (std.ctype.isdigit(pID[0]))
        {
            long lLevel = atoi(pID);
            if (vDebugLevel > mDebugLevel)
                return lLevel <= vDebugLevel ? true : false ;
            else
                return lLevel <= mDebugLevel ? true : false ;
        }
        else
            return (pID in vActiveDebugs) != null ? true :
                     (pID in mActiveDebugs) != null ? true : false ;
    }

    // -------------------------------------------
    void ProcessTokens (char[] pFileText, inout int pPos, char[] pEndStmt = kCloseBrace)
    // -------------------------------------------
    {
        /* This scans the source text for specific tokens until the 'pEndStmt'
           token is found, or end of text.
        */
        char[] lCurToken;
        char[] lPrevToken;
        int lLastPos;

        lCurToken = "";
        lLastPos = pFileText.length-1;

        lblNextToken:
        while ((lPrevToken = lCurToken.dup, lCurToken = GetNextToken (pFileText, pPos)) !is null) {

            switch(lCurToken) {
                case kOpenBrace:
                    ProcessTokens (pFileText, pPos);
                    break;

                case "\\":
                    // Ignore escaped tokens.
                    if (pPos <= lLastPos )
                        pPos++;
                    break;

                case "\"":
                    // ** drop through **
                case "\'":
                    // Handle various string and character literals.
                    while ((pPos <= lLastPos) && (pFileText[pPos] != lCurToken[0]))
                        if (pFileText[pPos] == '\\')
                            pPos += 2;
                        else
                            pPos++;
                    pPos++;
                    break;

                case "`": // Skip over a raw string
                    while ( (pPos <= lLastPos) && (pFileText[pPos] != '`'))
                        pPos++;
                    pPos++;
                    break;

                case "r":
                    if ( (pPos <= lLastPos) && (pFileText[pPos] == '\"') )
                    {   // Skip over a raw string
                        pPos++;
                        do {pPos++;}
                            while ( (pPos <= lLastPos) && (pFileText[pPos] != '\"'));
                        pPos++;
                    }
                    break;

                case "main":
                case "WinMain":
                case "DllMain":
                    if ( (lPrevToken != "class") &&
                         (lPrevToken != "template")
                        ) {
                        doMain(pFileText, pPos, lCurToken);
                    }
                    break;

                case "version":
                    doVersion(pFileText, pPos);
                    break;

                case "debug":
                    doDebug(pFileText, pPos);
                    break;

                case "import":
                    doImport(pFileText, pPos);
                    break;

                case "module":
                    doModule(pFileText, pPos);
                    break;

                case "pragma":
                    doPragma(pFileText, pPos);
                    break;

                default:
                    if (lCurToken in AttributeBlocks) {
                        doAttribute(pFileText, pPos, lCurToken);

                    } else if (lCurToken == pEndStmt) {
                        break lblNextToken;
                    }
            } // end switch
        } // end while
    }

    //----------------------------------------------------------
    void doMain (in char[] pFileText, inout int pPos, char[] pCurrentToken)
    {
        int lSavedPos;
        char[] lToken;

        lSavedPos = pPos;
        lToken = GetNextToken(pFileText,pPos);
        if (lToken == kOpenParen) {
            mainFound = true;
            version(Windows) {
                if (pCurrentToken == "WinMain") {
                    mainGUI = true;
                }
                else if (pCurrentToken == "DllMain") {
                    mainDLL = true;
                }
            }
        } else {
            pPos = lSavedPos;
        }

    }

    //----------------------------------------------------------
    void doVersion (in char[] pFileText, inout int pPos)
    {
        char[] lCurrentToken;
        char[] lVersionId;
        int lSavedPos;

        lCurrentToken = GetNextToken (pFileText, pPos);
        if (lCurrentToken == "=")
        {
            // Activate the following identifer version.
            this.ActivateVersion( GetNextToken (pFileText, pPos) );
            GetNextToken(pFileText, pPos); // Throw away the trailing ';'
            return;
        }

        if (lCurrentToken != kOpenParen)
        {
            return; // Not likely as this is bad syntax.
        }

        // **** Assume that correct syntax has been used. ****
        // Pick out the version identifier and skip over the closing parenthesis.
        lVersionId = GetNextToken (pFileText, pPos);
        lCurrentToken = GetNextToken (pFileText, pPos);

        if ( IsActiveVersion(lVersionId) )
        {
            // thus the following tokens must be processed.
            lSavedPos = pPos; // Remember where we are.
            lCurrentToken = GetNextToken (pFileText, pPos);
            if (lCurrentToken == kOpenBrace)
            {
                // process the stuff in the block
                ProcessTokens (pFileText, pPos);
            }
            else
            {
                // Not a block so we just process the next statement only.
                pPos = lSavedPos;  // Back up one token.
                ProcessTokens (pFileText, pPos, ";");
            }

            /* If the next token is an 'else', I must skip over the statement
               or block that follows the 'else'. Otherwise this marks the
               end of the version statement.
            */
            // Remember where we are.
            lSavedPos = pPos;
            lCurrentToken = GetNextToken(pFileText, pPos);
            if (lCurrentToken == "else")
            {
                lCurrentToken = GetNextToken(pFileText, pPos);
                if (lCurrentToken == kOpenBrace)
                {
                    // Skip over a block
                    skipContent(pFileText, pPos);
                }
                else
                {
                    // skip over a statement.
                    skipContent(pFileText, pPos, ";");
                }
            }
            else
            {
                // It wasn't an 'else' so be back up and let the calling
                // routine handle it.
                pPos = lSavedPos;
            }
            return;
        }

        /* Not an active version so therefore the following statement/block
           must not be processed.
        */
        lCurrentToken = GetNextToken(pFileText, pPos);

        if ( lCurrentToken != kOpenBrace)
        {
            skipContent(pFileText, pPos, ";");
            return;
        }

        skipContent(pFileText, pPos);

        lSavedPos = pPos;
        lCurrentToken = GetNextToken(pFileText, pPos);
        if (lCurrentToken != "else")
        {
            pPos = lSavedPos;
            return;
        }

        lSavedPos = pPos;
        lCurrentToken = GetNextToken(pFileText, pPos);
        if (lCurrentToken == kOpenBrace)
        {
            ProcessTokens(pFileText, pPos);
        }
        else
        {
            pPos = lSavedPos;
            ProcessTokens(pFileText, pPos, ";");
        }

    }

    //----------------------------------------------------------
    void doDebug(in char[] pFileText, inout int pPos)
    {
        char[] lCurrentToken;
        char[] lDebugId;
        int lSavedPos;

        lSavedPos = pPos;
        lCurrentToken = GetNextToken (pFileText, pPos);
        if (lCurrentToken == "=")
        {
            // Activate the following debug identifer.
            this.ActivateDebug( GetNextToken (pFileText, pPos) );
            GetNextToken(pFileText, pPos); // Throw away the trailing ';'
            return;
        }

        if (lCurrentToken != kOpenParen)
        {
            lDebugId = "1";
            // Back track to the token after the 'debug'
            pPos = lSavedPos;
        }
        else
        {
            // Pick out the debug identifier and skip over the closing parenthesis.
            lDebugId = GetNextToken (pFileText, pPos);
            lCurrentToken = GetNextToken (pFileText, pPos);
        }

        if ( IsActiveDebug(lDebugId) )
        {
            // thus the following tokens must be processed.
            lSavedPos = pPos; // Remember where we are.
            lCurrentToken = GetNextToken (pFileText, pPos);
            if (lCurrentToken == kOpenBrace)
            {
                // process the stuff in the block
                ProcessTokens (pFileText, pPos);
            }
            else
            {
                // Not a block so we just process the next statement only.
                pPos = lSavedPos;  // Back up one token.
                ProcessTokens (pFileText, pPos, ";");
            }

            /* If the next token is an 'else', I must skip over the statement
               or block that follows the 'else'. Otherwise this marks the
               end of the debug statement.
            */
            // Remember where we are.
            lSavedPos = pPos;
            lCurrentToken = GetNextToken(pFileText, pPos);
            if (lCurrentToken == "else")
            {
                lCurrentToken = GetNextToken(pFileText, pPos);
                if (lCurrentToken == kOpenBrace)
                {
                    // Skip over a block
                    skipContent(pFileText, pPos);
                }
                else
                {
                    // skip over a statement.
                    skipContent(pFileText, pPos, ";");
                }
            }
            else
            {
                // It wasn't an 'else' so be back up and let the calling
                // routine handle it.
                pPos = lSavedPos;
            }
            return;
        }

        /* Not an active version so therefore the following statement/block
           must not be processed.
        */
        lCurrentToken = GetNextToken(pFileText, pPos);

        if ( lCurrentToken != kOpenBrace)
        {
            skipContent(pFileText, pPos, ";");
            return;
        }

        skipContent(pFileText, pPos);

        lSavedPos = pPos;
        lCurrentToken = GetNextToken(pFileText, pPos);
        if (lCurrentToken != "else")
        {
            pPos = lSavedPos;
            return;
        }

        lSavedPos = pPos;
        lCurrentToken = GetNextToken(pFileText, pPos);
        if (lCurrentToken == kOpenBrace)
        {
            ProcessTokens(pFileText, pPos);
        }
        else
        {
            pPos = lSavedPos;
            ProcessTokens(pFileText, pPos, ";");
        }

    }

    //----------------------------------------------------------
    void doImport (in char[] pFileText, inout int pPos)
    {
        char[] lCurrentToken;
        char[] lModName;
        int lSavedPos;

        // import lModName.lModName, lModName, lModName.lModName.lModName;

        while ((lSavedPos = pPos,
                lCurrentToken = GetNextToken (pFileText, pPos)) !is null
                   && (lCurrentToken != ";"))
        {
            if ( (lCurrentToken == ",") && (lModName.length > 0)) {
                mReferencedImports ~= ModuleToFilename(lModName);
                lModName = "";
            } else {
                lModName ~= lCurrentToken;
            }
        }
        // Don't forget the last one!
        if (lModName.length > 0) {
            mReferencedImports ~= ModuleToFilename(lModName);
        }
        if (lCurrentToken == ";")
            pPos = lSavedPos;
    }
    //----------------------------------------------------------
    void doModule (in char[] pFileText, inout int pPos)
    {
        char[] lCurrentToken;
        char[] lModName;
        int lSavedPos;

        // module lModName [.lModName];
        while ((lSavedPos = pPos,
                lCurrentToken = GetNextToken (pFileText, pPos)) !is null
                   && (lCurrentToken != ";"))
        {
            lModName ~= lCurrentToken;
        }

        if (lModName.length > 0 )
        {
            mModuleName = lModName;
            if (mVerboseMode)
            {
                writefln("Module name set to '%s'", mModuleName);
            }
        }

        if (lCurrentToken == ";")
            pPos = lSavedPos;
    }

    //----------------------------------------------------------
    void doPragma (in char[] pFileText, inout int pPos)
    {
        /* Looking for syntax form ... pragma(<type> [ , <id> [,<id>]...] );
        */
        char[] lCurrentToken;
        char[] lPragmaId;
        int lSavedPos;

        lPragmaId.length = 0;
        if (GetNextToken(pFileText, pPos) == kOpenParen )
        {
            lCurrentToken = GetNextToken(pFileText,pPos);
            if (lCurrentToken == "link")
            {
                while(true)
                {
                    lPragmaId = "";
                    while ( (lCurrentToken = GetNextToken(pFileText,pPos)) !is null)
                    {
                        if ( (lCurrentToken == "\"") || (lCurrentToken == "`") ) {
                            lCurrentToken = GetStringLit(pFileText,pPos, lCurrentToken[0]);
                        }

                        if ((lCurrentToken != ",") && (lCurrentToken != kCloseParen))
                        {
                            lPragmaId ~= lCurrentToken;
                        }
                        else
                            break;
                    }

                    // If it has no extension, add the 'lib' one.
                    for(int i = lPragmaId.length-1; i >=0 ; i--)
                    {
                        if (lPragmaId[i] == '.')
                        {
                            // Check to see if this is a 'versioned'
                            // library file name, eg.  "glade-2.0"
                            if (i == lPragmaId.length - 1)
                                lPragmaId ~= LibExt;
                            else if (std.ctype.isdigit(lPragmaId[i+1]))
                                lPragmaId ~= "." ~ LibExt;
                            break;
                        }
                        if (i == 0)
                        {
                            // Not a dot to be seen ;-)
                            lPragmaId ~= "." ~ LibExt;
                        }
                    }
                    // Add link path to compiler switches.
                    if (AddLink != null)
                        AddLink(lPragmaId);
                    if (lCurrentToken == kCloseParen)
                    {
                        break;
                    }
                }
            } else if (lCurrentToken == "build_def"){
                // Collect records for the optlink module definition file.
                lPragmaId = "";
                while ( (lCurrentToken = GetNextToken(pFileText,pPos)) !is null)
                {
                    if ( (lCurrentToken == "\"") || (lCurrentToken == "`") ) {
                        lCurrentToken = GetStringLit(pFileText,pPos, lCurrentToken[0]);
                    }

                    if ((lCurrentToken != ",") && (lCurrentToken != kCloseParen)) {
                        if (AddBuildDef != null)
                            AddBuildDef(lCurrentToken);
                    }
                    if (lCurrentToken == kCloseParen){
                        break;}
                }

            } else if (lCurrentToken == "nolink"){
                // This source file is NOT to be linked
                mNoLink = true;

                // Skip over the trailing parenthesis.
                GetNextToken(pFileText,pPos);

            } else if (lCurrentToken == "target"){
                if ( (lCurrentToken = GetNextToken(pFileText,pPos)) == ",")
                {
                    lCurrentToken = GetNextToken(pFileText,pPos);
                    if ( (lCurrentToken == "\"") || (lCurrentToken == "`") ) {
                        lCurrentToken = GetStringLit(pFileText,pPos, lCurrentToken[0]);
                    }
                    if (AddTarget != null)
                        AddTarget(lCurrentToken);
                }

                while(lCurrentToken != kCloseParen)
                {   // Skip everything until we find a closing paren.
                    lCurrentToken = GetNextToken(pFileText,pPos);
                }

            } else if (lCurrentToken == "build"){
                if ( (lCurrentToken = GetNextToken(pFileText,pPos)) == ",")
                {
                    char[] lExternFile;
                    char[][] lExternOpts;

                    lExternFile = GetNextToken(pFileText,pPos);
                    if ( (lExternFile == "\"") || (lExternFile == "`") ) {
                        lExternFile = GetStringLit(pFileText,pPos, lExternFile[0]);
                    }

                    while ((lCurrentToken = GetNextToken(pFileText,pPos)) == ",")
                    {
                        lExternOpts.length = lExternOpts.length + 1;
                        lExternOpts[$-1] = GetNextToken(pFileText,pPos);
                        if (lExternOpts[$-1] == "\"") {
                            lExternOpts[$-1] = GetStringLit(pFileText,pPos);
                        }
                    }

                    if (AddExternal != null)
                        AddExternal(lExternFile, lExternOpts);
                }

                while(lCurrentToken != kCloseParen)
                {   // Skip everything until we find a closing paren.
                    lCurrentToken = GetNextToken(pFileText,pPos);
                }

            } else if (lCurrentToken == "include"){

                if ( (lCurrentToken = GetNextToken(pFileText,pPos)) == ",")
                {
                    lCurrentToken = GetNextToken(pFileText,pPos);
                    if ( (lCurrentToken == "\"") || (lCurrentToken == "`") ) {
                        lCurrentToken = GetStringLit(pFileText,pPos, lCurrentToken[0]);
                    }
                mReferencedImports ~= ModuleToFilename(lCurrentToken);
                }

                while(lCurrentToken != kCloseParen)
                {   // Skip everything until we find a closing paren.
                    lCurrentToken = GetNextToken(pFileText,pPos);
                }

            } else if (lCurrentToken == "export_version"){

                while ( (lCurrentToken = GetNextToken(pFileText,pPos)) == ",")
                {
                    lCurrentToken = GetNextToken(pFileText,pPos);
                    if ( (lCurrentToken == `"`) || (lCurrentToken == "`") )
                    {
                        lCurrentToken = `"` ~ GetStringLit(pFileText,pPos, lCurrentToken[0]) ~ `"`;
                    }
                    AddCompilerArg(`+v+` ~ lCurrentToken );
                }

                while(lCurrentToken != kCloseParen)
                {   // Skip everything until we find a closing paren.
                    lCurrentToken = GetNextToken(pFileText,pPos);
                }
            }

        }
    }

    //----------------------------------------------------------
    void doAttribute (in char[] pFileText, inout int pPos, char[] pAttrType)
    {
        char[] lCurrentToken;
        char[] lNextToken;
        int lSavedPos;

        // Handle private/public/etc... blocks.
        lSavedPos = pPos;
        lCurrentToken = GetNextToken(pFileText,pPos);

        if (lCurrentToken == kOpenBrace) // read a block statement
        {
            ProcessTokens (pFileText, pPos);

        } else {
            // Check for special auto build number import.
            if ((pAttrType == "private") && (lCurrentToken == "import"))
            {
                lCurrentToken = GetNextToken(pFileText,pPos);
                while ( (lNextToken = GetNextToken(pFileText,pPos)) != ";")
                {
                    lCurrentToken ~= lNextToken;
                }

                if (lCurrentToken == mModuleName ~ "_bn")
                {
                    mBuildNumber = LoadBuildNumber(mModuleName, mFileName);
                }

            }
            // Revert to the position just after the attribute.
            pPos = lSavedPos;
        }

    }

    //----------------------------------------------------------
    /* Get the next token, skipping over comments. */
    char[] GetNextToken (in char[] pFileText, inout int pPos)
    {
        int lLastPos;

        lLastPos = pFileText.length - 1;
        while (pPos <= lLastPos)
        {
            // Check for comments ...
            if ( (pFileText[pPos] == '/') && (pPos < lLastPos) )
            {
                // Check for multiline unnested comments
                if (pFileText [pPos+1] == '*') {
                    pPos += 2;
                    while (pPos < lLastPos) {
                        if (pFileText [pPos] == '*' && pFileText [pPos+1] == '/') {
                            pPos += 2;
                            break;
                        }
                        else
                            pPos++;
                    }
                } else
                // Check for single line comments
                if (pFileText [pPos+1] == '/') {
                    pPos += 2;
                    while ((pPos <= lLastPos) && (pFileText[pPos] != '\n')) {
                        pPos++;
                        }

                } else
                // Check for multiline nested comments
                if (pFileText [pPos+1] == '+') {
                    int lDepth = 1;

                    pPos += 2;
                    while (pPos <= lLastPos) {
                        if (pFileText [pPos] == '/'){
                            if ((pPos < lLastPos) && (pFileText [pPos+1] == '+')) {
                                lDepth++;
                                pPos += 2;
                            }
                            else
                                pPos++;
                        } else if (pFileText [pPos] == '+') {
                            if ((pPos < lLastPos) && (pFileText [pPos+1] == '/')) {
                                pPos += 2;
                                if (lDepth == 1) {
                                    break;
                                }
                                lDepth--;
                            }
                            else
                                pPos++;
                        } else {
                            pPos++;
                        }
                    }

                    if (pPos > lLastPos) { throw new Exception("Mismatched nested comments in " ~ mFileName); }
                }
            } // else
            // Check for 'words'
            if (isalpha (pFileText [pPos])) {
                int lStart = pPos;
                pPos++;
                while ((pPos <= lLastPos) && ((isalnum (pFileText [pPos]) || pFileText [pPos] == '_'))) {
                    pPos ++;
                }
                return pFileText[lStart .. pPos];

            } else
            // Check for whitespace
            if (isspace (pFileText [pPos])) {
                pPos++;

            } else
            {
            // Return whatever character we have discovered.
                pPos++;
                return pFileText[pPos-1 .. pPos];
            }
        }

        return null;
    }

    //----------------------------------------------------------
    void skipContent (in  char[] pFileText, inout int pPos, char[] pEndStmt = kCloseBrace)
    {
        // Skips thru nested blocks.

        /* The initial token is assumed to be
           the first token inside a '{' '}' pair.
           */
        char[] lCurrentToken;

        while ((lCurrentToken = GetNextToken (pFileText, pPos)) !is null)
        {
            if (lCurrentToken == pEndStmt)
            {
                return;
            }
            else if (lCurrentToken == kOpenBrace)
                skipContent (pFileText, pPos);

        }
    }

    //----------------------------------------------------------
    char[] GetStringLit (in  char[] pFileText, inout int pPos, char pEndStmt = '"')
    {
        int lStartLit;
        int lEndLit;

        lStartLit = pPos;

        while( (pPos < pFileText.length) && (pFileText[pPos] != pEndStmt) )
        {
            pPos++;
        }
        lEndLit = pPos;
        if (pPos < pFileText.length)
            pPos++; // Skip over lit end char.

        return pFileText[lStartLit .. lEndLit];

    }


  } // end private

} // end class

char[] ModuleToFilename(char[] pModuleName)
{
    char[] lFileName;

    // Replace dots with the opsys path separator
    lFileName = util.str.ReplaceChar(pModuleName.dup, '.', std.path.sep[0]);
    // Add the 'd' extention and append the full path
    lFileName = GetFullPathname( addExt( lFileName, "d" ) );

    if(mVerboseMode) {
        writefln(" module->file %s => %s",pModuleName,lFileName);
    }
    return lFileName;
}

char[] FileToModulename(char[] pFileName)
{
    char[] lModuleName;
    char[] lTemp;

    // Copy the name
    lModuleName = util.pathex.AbbreviateFileName(pFileName, GetImportRoots() );


    // Remove file extension.
    if (lModuleName[$-2 .. $] == ".d")
    {
        lModuleName.length = lModuleName.length - 2;
    }

    version(Windows) {
        // Remove the 'Drive' letter if present.
        if (lModuleName.length > 1  &&  lModuleName[1] == ':')
        {
            lModuleName = lModuleName[2..$];
        }
        if (lModuleName.length > 0  &&  lModuleName[0] == std.path.sep[0])
        {
            lModuleName = lModuleName[1..$];
        }
    }
    // Replace opsys path separators with dots.
    lModuleName = util.str.ReplaceChar(lModuleName, std.path.sep[0], '.');

    if(mVerboseMode) {
        writefln(" file->module %s => %s",pFileName,lModuleName);
    }
    return lModuleName;
}


int LoadBuildNumber(char[] pModuleName, char[] pFileName)
{
    char[][] lFileLines;
    char[] lFileName;
    int lBuildNumber = 0;
    char[] lBaseDir;

    lFileName = ModuleToFilename(pModuleName ~ "_bn");
    lBaseDir = std.path.getDirName(pFileName);
    if (lBaseDir.length > 0 && std.path.getDirName(lFileName).length == 0)
        lFileName = std.path.getDirName(pFileName) ~ std.path.sep ~ lFileName;

    lFileName = util.pathex.AbbreviateFileName (
                    util.pathex.CanonicalPath(lFileName, false));
    if (std.file.exists(lFileName))
    {
        lFileLines = GetTextLines( lFileName, util.fileex.GetOpt.Exists );
        BLK_FindBN:
        foreach (char[] lLine; lFileLines)
        {
            if ( util.str.IsLike(lLine, cast(dchar[])"*long auto_build_number = *;*") )
            {
                for(int i = find(lLine,"=") + 2; i < lLine.length; i++)
                {
                    int n = find("0123456789", lLine[i..i+1] );
                    if (n >= 0)
                        lBuildNumber = lBuildNumber * 10 + n;
                    else if(lLine[i] == ';')
                        break;
               }
            }
        }
    }
    else
    {
        std.file.write(lFileName,
            std.string.format("module %s;\n" , FileToModulename(lFileName)) ~
            "// This file is automatically maintained by the BUILD utility,\n"
            "// Please refrain from manually editing it.\n"
            "long auto_build_number = 0;\n"
            );
        lBuildNumber = 0;
    }
    return lBuildNumber;
}

// -------------------------------------------
void ActivateVersion(char[] pID)
// -------------------------------------------
{
    if (pID.length > 0)
    {
        if (std.ctype.isdigit (pID[0]))
        // Note that even though we capture this here,
        // levels are not yet implemented in the
        // source scanner.
            vVersionLevel = atoi(pID);
        else
            vActiveVersions [ pID ] = true;
    }
}

// -------------------------------------------
void ActivateDebug(char[] pID)
// -------------------------------------------
{
    static const char[] lDefaultLevel = "1";
    if (pID.length == 0)
        pID = lDefaultLevel;

    if (std.ctype.isdigit(pID[0]))
    // Note that even though we capture this here,
    // levels are not yet implemented in the
    // source scanner.
        vDebugLevel = atoi(pID);
    else
        vActiveDebugs [ pID ] = true;

}

// -------------------------------------------
void SetKnownVersions()
// -------------------------------------------
{

    version(DigitalMars)  ActivateVersion("DigitalMars");
    version(X86)          ActivateVersion("X86");
    version(PPC)          ActivateVersion("PPC");
    version(AMD64)        ActivateVersion("AMD64");
    version(PPC64)        ActivateVersion("PPC64");
    version(Windows)      ActivateVersion("Windows");
    version(Win32)        ActivateVersion("Win32");
    version(Win64)        ActivateVersion("Win64");
    version(linux)        ActivateVersion("linux");
    version(darwin)       ActivateVersion("darwin");
    version(Unix)         ActivateVersion("Unix");
    version(unix)         ActivateVersion("unix");
    version(Posix)        ActivateVersion("Posix");
    version(LittleEndian) ActivateVersion("LittleEndian");
    version(BigEndian)    ActivateVersion("BigEndian");
    version(D_InlineAsm)  ActivateVersion("D_InlineAsm");

    if (mVerboseMode)
        foreach(char[] k; vActiveVersions.keys){
            writefln("Active Version: '%s'", k);
        }
}
