/**************************************************************************

        @file fileex.d

        Copyright (c) 2005 Derek Parnell

        This software is provided 'as-is', without any express or implied
        warranty. In no event will the authors be held liable for damages
        of any kind arising from the use of this software.

        Permission is hereby granted to anyone to use this software for any
        purpose, including commercial applications, and to alter it and/or
        redistribute it freely, subject to the following restrictions:

        1. The origin of this software must not be misrepresented; you must
           not claim that you wrote the original software. If you use this
           software in a product, an acknowledgment within documentation of
           said product would be appreciated but is not required.

        2. Altered source versions must be plainly marked as such, and must
           not be misrepresented as being the original software.

        3. This notice may not be removed or altered from any distribution
           of the source.

        4. Derivative works are permitted, but they must carry this notice
           in full and credit the original source.


                        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


        @version        Initial version, March 2005
        @author         Derek Parnell


**************************************************************************/

module util.fileex;
private import util.fileex_bn;   // The build number for this module.

private{
    import std.file;
    import std.string;
    import std.path;
    import std.ctype;

    debug(1)
    {
        import std.stdio;
    }



}

// Module constructor
// ----------------------------------------------
static this()
// ----------------------------------------------
{
    debug(1)
    {
        writefln("util.fileex build #%d", auto_build_number);
    }
}

enum GetOpt
{
    Exists = 'e',   // Must exist otherwise Get fails.
    Always = 'a'    // Get always returns something, even if it's just
                    //   empty lines for a missing file.
};

// Read an entire file into a string.
char[] GetText(char[] pFileName, GetOpt pOpt = GetOpt.Always)
{
    char[] lFileText;

    if (std.file.exists( pFileName))
    {
        lFileText = cast(char[]) std.file.read(pFileName);
        if ( (lFileText.length == 0) ||
             (lFileText[$-1] != '\n'))
             lFileText ~= std.path.linesep;
    }
    else if (pOpt == GetOpt.Exists) {
        throw new Exception( std.string.format("File '%s' not found.", pFileName));
    }
    return lFileText;

}

// Read a entire file in to a set of lines (strings).
char[][] GetTextLines(char[] pFileName, GetOpt pOpt = GetOpt.Always)
{
    char[][] lLines;
    char[]   lText;

    lText = GetText(pFileName, pOpt);
    lLines = std.string.splitlines( lText );
    return lLines;
}

enum CreateOpt
{
    New = 'n',      // Must create a new file, thus it cannot already exist.
    Create = 'c',   // Can either create or replace; it doesn't matter.
    Replace = 'r'   // Must replace an existing file.
};

void CreateTextFile(char[] pFileName, char[][] pLines, CreateOpt pOpt = CreateOpt.Create)
{
    char[] lBuffer;
    bool lFileExists;

    lFileExists = (std.file.exists( pFileName) ? true : false);
    if (pOpt == CreateOpt.Replace && !lFileExists)
        throw new Exception( std.string.format("File '%s' doesn't exist.", pFileName));

    if (pOpt == CreateOpt.New && lFileExists)
        throw new Exception( std.string.format("File '%s' already exists.", pFileName));

    if (std.file.exists(pFileName))
        std.file.remove(pFileName);

    foreach(char[] lText; pLines) {
        // Strip off any trailing line-end chars.
        for(int i = lText.length-1; i >= 0; i--)
        {
            if (find(std.path.linesep, lText[i]) == -1)
            {
                if (i != lText.length-1)
                    lText.length = i+1;
                break;
            }
        }

        // Append the opsys' line-end convention.
        lBuffer ~= lText ~ std.path.linesep;
    }
    std.file.write(pFileName, lBuffer);
}

void CreateTextFile(char[] pFileName, char[] pLines, CreateOpt pOpt = CreateOpt.Create)
{
    char[] lBuffer;
    bool lFileExists;
    char[][] lLines;

    // Split into lines, disregarding line-end conventions.
    lLines = std.string.splitlines(pLines);
    // Write out the text using the opsys' line-end convention.
    CreateTextFile(pFileName, lLines, pOpt);
}

ulong[] FindInFile(char[] pFileName, char[] pText, char[] pOptions = "", uint pMax=1)
{
    bool lCaseSensitive;
    bool lRegExp;
    bool lWordOnly;
    bool lCounting;
    auto char[] lBuffer;
    int lPos;
    ulong[] lResult;

    // TODO //
    // Implement options. ---------------------------------- //

    lCaseSensitive = true;
    if (find(pOptions, 'i') != -1)
        lCaseSensitive = true;
    if (find(pOptions, 'I') != -1)
        lCaseSensitive = true;

    lRegExp = false;
    if (find(pOptions, 'r') != -1)
        lRegExp = true;
    if (find(pOptions, 'R') != -1)
        lRegExp = true;

    lWordOnly = false;
    if (find(pOptions, 'w') != -1)
        lWordOnly = true;
    if (find(pOptions, 'W') != -1)
        lWordOnly = true;

    lCounting = true;
    if (find(pOptions, 'a') != -1)
        lCounting = false;
    if (find(pOptions, 'A') != -1)
        lCounting = false;

    lBuffer = cast(char[])std.file.read(pFileName);
    lResult.length = 0;
    lPos = 0;

    while ( (lPos = find(lBuffer[lPos..$], pText)) != -1)
    {
        if (lWordOnly)
        {
            if (lPos > 0)
            {
                if (std.ctype.isalnum(lBuffer[lPos-1]) )
                {
                    lPos++;
                    continue;
                }
            }
            if (lPos + pText.length < lBuffer.length)
            {
                if (std.ctype.isalnum(lBuffer[lPos + pText.length - 1]) )
                {
                    lPos++;
                    continue;
                }
            }
        }

        lResult.length = lResult.length + 1;
        lResult[$-1] = lPos;
        if (lCounting)
        {
            pMax--;
            if (pMax == 0)
                break;
        }

        lPos += pText.length;
        if ((lBuffer.length - lPos) < pText.length)
            break;
    }
    return lResult;
}

private bool[char[]] lExistingFiles;

// --------------------------------------------------
bool FileExists(char[] pFileName)
{
    if (pFileName in lExistingFiles)
        return true;
    if(std.file.exists(pFileName))
    {
        lExistingFiles[pFileName] = true;
        return true;
    }
    return false;
}

// --------------------------------------------------
void PurgeFileExistsCache()
{
    char[][] lKeys;

    lKeys = lExistingFiles.keys.dup;
    foreach(char[] lFile; lKeys)
    {
        lExistingFiles.remove(lFile);
    }
}